(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

{
open Lang_common
open Lang_src
open Parser
}
rule token = parse
  | [' ' '\t' '\n']  { token lexbuf }
  | eof { EOF }
  | ';'  { SEMICOLON }
  | '('  { LPAREN }
  | ')'  { RPAREN }
  | ['0'-'9']+ as lxm  { NUM(Big_int.big_int_of_string lxm) }
  | '+'  { BINOP(Plus) }
  | '-'  { BINOP(Minus) }
  | '*'  { BINOP(Mult) }
  | '='  { BINOP(Equal) }
  | "input"  { INPUT }
  | "output"  { OUTPUT }
  | "var"  { VAR }
  | "lbl"  { LBL }
  | "unit"  { UNIT }
  | "nat"  { NAT }
  | "bop"  { BOP }
  | "ifnz"  { IFNZ }
  | "pair"  { PAIR }
  | "fst"  { FST }
  | "snd"  { SND }
  | "inl"  { INL }
  | "inr"  { INR }
  | "case"  { CASE }
  | "roll"  { ROLL }
  | "fix"  { FIX }
  | "app"  { APP }
  | "gen"  { GEN }
  | "inst"  { INST }
  | "pack"  { PACK }
  | "unpack"  { UNPACK }
  | "new"  { NEW }
  | "read"  { READ }
  | "write"  { WRITE }
  | "refeq"  { REFEQ }
  | "let"  { LET }
