(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw_common gw ssr util ts.
Require Import src2tgt src2tgt_correct_aux.

Unset Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Lemma tr_exp_correct_case_inl: forall  
  (F : list (label * typ))
  (G : list (termvar * typ))
  (tyl : typ)
  (tyr : typ)
  (ty : typ)
  (e : exp)
  (x : termvar)
  (el : exp)
  (er : exp)
  (i : wlift mylw)
  (i0 : wlift mylw)
  (IHl : ts_psim_aux i0 F ((x, tyl) :: G) ty
          (fun n : word => map enc (tr_exp el n (map fst F) (x :: map fst G)))
          el)
  (i1 : wlift mylw)
  (IHr : ts_psim_aux i1 F ((x, tyr) :: G) ty
          (fun n : word => map enc (tr_exp er n (map fst F) (x :: map fst G)))
          er)
  (tsub : typename -> typ)
  (g : state (wlift mylw) -> vrelf)
  (r : regfile)
  (P : tquery_state -> lang_spec.val tgt_lsi -> Prop)
  (s0 : state src_gw_sts)
  (s1 : state ref_sts)
  (s2 : state mylw)
  (r0 : regfile)
  (P0 : tquery_state -> lang_spec.val tgt_lsi -> Prop)
  (s : state src_gw_sts)
  (s3 : state ref_sts)
  (s4 : state mylw)
  (LCT : lc_tsub tsub)
  (GK : gknow (wlift mylw) g)
  (EXT : ext (wlift mylw) (r0, P0, s, s3, s4) (r, P, s0, s1, s2))
  (senva : list (label * lang_spec.val tgt_lsi))
  (RSV : forall (l : label) (ty' : typ),
        lookup l F = Some ty' ->
        exists (va : lang_spec.val tgt_lsi) (vb : val),
          lookup l senva = Some va /\
          lookup l s = Some vb /\
          vclos (wlift mylw) (r0, P0, s, s3, s4) (g (r0, P0, s, s3, s4))
            (tsubst_name tsub ty') va vb)
  (SORD : map fst F = map fst senva)
  (denva : list (termvar * lang_spec.val tgt_lsi))
  (denvb : termvar -> lang_spec.val src_lsi)
  (RDV : forall (x : termvar) (ty' : typ),
        lookup x G = Some ty' ->
        exists va : lang_spec.val tgt_lsi,
          lookup x denva = Some va /\
          vclos (wlift mylw) (r0, P0, s, s3, s4) (g (r0, P0, s, s3, s4))
            (tsubst_name tsub ty') va (denvb x))
  (ENV : reprlist (snd s4) (r0 r_env) (map snd denva))
  (DORD : map fst G = map fst denva)
  (CL : forall x : termvar,
       In x (map fst G) -> forall y : termvar, fv_exp (denvb x) y -> False)
  (WCL : subst_weakly_closed (fun x : termvar => denvb x))
  (laddr : word)
  (n : word)
  (NZ : laddr <> 0)
  (h : pmap (lang_spec.val tgt_lsi))
  (Eladdr : snd s4 =
           {laddr |=> map enc (tr_exp (Ecase e x el er) n (map fst F) (map fst G))}
           ## {laddr - n |=> map snd senva} ## h)
  (j : wlift mylw)
  (kia : lang_spec.cont tgt_lsi)
  (kib : lang_spec.cont src_lsi)
  (kb : cont)
  (ty' : typ)
  (p : pmvalid (snd s4))
  (g' : state (wlift mylw) ->
        flextyp -> lang_spec.val tgt_lsi -> lang_spec.val src_lsi -> Prop)
  (s' : state (wlift mylw))
  (va : word)
  (vb : lang_spec.val src_lsi)
  (LEG : forall (s : state (wlift mylw)) (ty : flextyp)
           (va : lang_spec.val tgt_lsi) (vb : lang_spec.val src_lsi),
         g s ty va vb -> g' s ty va vb)
  (GK' : gknow (wlift mylw) g')
  (EXT0 : ext (wlift mylw) s' (r0, P0, s, s3, s4))
  (ENV0 : fst (fst (fst (fst s'))) r_env = r0 r_env)
  (SP : fst (fst (fst (fst s'))) r_sp = r0 r_sp + 1)
  (RET : fst (fst (fst (fst s'))) r_ret = r0 r_ret)
  (LST : fst (snd s') = {fst (fst (fst (fst s'))) r_sp - 1 |-> va} ## fst s4)
  (mb : mach src_lsi)
  (KSIM : forall
           (g' : state (wlift mylw) ->
                 flextyp ->
                 lang_spec.val tgt_lsi -> lang_spec.val src_lsi -> Prop)
           (s' : state (wlift mylw)) (va : word) (vb : lang_spec.val src_lsi),
         (forall (s : state (wlift mylw)) (ty : flextyp)
            (va0 : lang_spec.val tgt_lsi) (vb0 : lang_spec.val src_lsi),
          g s ty va0 vb0 -> g' s ty va0 vb0) ->
         gknow (wlift mylw) g' ->
         ext (wlift mylw) s' (r0, P0, s, s3, s4) ->
         fst (fst (fst (fst s'))) r_env = r0 r_env ->
         fst (fst (fst (fst s'))) r_sp = r0 r_sp + 1 ->
         fst (fst (fst (fst s'))) r_ret = r0 r_ret ->
         fst (snd s') = {fst (fst (fst (fst s'))) r_sp - 1 |-> va} ## fst s4 ->
         vclos (wlift mylw) s' (g' s') (tsubst_name tsub ty) va vb ->
         esim' (wlift mylw) j kia kib g' (r, P, s0, s1, s2) s' ty'
           {|
           cpc := Just
                    (length
                       (map enc
                          (tr_exp (Ecase e x el er) n (map fst F) (map fst G))) +
                     laddr);
           crf := Nothing;
           cst := pmemp;
           chp := pmemp |} (pmemp, Nothing, Just (app_cont kb vb)))
  (refha : heap)
  (refhb : lang_src.heap)
  (freest : stack)
  (hp : heap)
  (CPa : tgt_confpred (fst (fst (fst s')))
          {|
          cpc := Nothing;
          crf := Just (fst (fst (fst (fst s'))));
          cst := freest;
          chp := hp |})
  (CPb : src_confpred (proj_env s') (pmemp, Just (proj_env s'), Nothing))
  (WHY0 : ts_refconfrel (fun sg : state ts_gw => g' (sg, snd s')) 
           (fst s')
           {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := refha |}
           (refhb, Nothing, Nothing))
  (RLb : real src_lsi
          (refhb, Just (proj_env s'),
          Just
            (app_cont kb
               (Ecase vb x
                  (subst_exp [Evar x // x]>> (fun x : termvar => denvb x) el)
                  (subst_exp [Evar x // x]>> (fun x : termvar => denvb x) er))))
          mb)
  (SV : pmvalid (freest ## fst (snd s')))
  (FIN : pmfinite (hp ## refha ## snd (snd s')))
  (HV : pmvalid (hp ## refha ## snd (snd s')))
  (RLa0 : real tgt_lsi
           {|
           cpc := Just
                    (length (map enc (tr_exp e n (map fst F) (map fst G))) +
                     laddr);
           crf := Just (fst (fst (fst (fst s'))));
           cst := freest ## fst (snd s');
           chp := hp ## refha ## snd (snd s') |}
           {|
           mpc := length (map enc (tr_exp e n (map fst F) (map fst G))) +
                  laddr;
           mrf := fst (fst (fst (fst s')));
           mst := freest ## fst (snd s');
           mhp := hp ## refha ## snd (snd s') |})
  (vax : lang_spec.val tgt_lsi)
  (vbx : lang_spec.val src_lsi)
  (CA : vquerya (wlift mylw) s' (valid_inl vax) va)
  (CB : vqueryb (wlift mylw) s' (valid_inl vbx) vb)
  (VRX : vclos (wlift mylw) s' (g' s') (tsubst_name tsub tyl) vax vbx),
   pesim bote (wfo_add i j) kia kib g' (r, P, s0, s1, s2) s' false ty'
     {|
     cpc := Just
              (length (map enc (tr_exp e n (map fst F) (map fst G))) + laddr);
     crf := Just (fst (fst (fst (fst s'))));
     cst := freest ## fst (snd s');
     chp := hp ## refha ## snd (snd s') |}
     (refhb, Just (proj_env s'),
     Just
       (app_cont kb
          (Ecase vb x
             (subst_exp [Evar x // x]>> (fun x0 : termvar => denvb x0) el)
             (subst_exp [Evar x // x]>> (fun x0 : termvar => denvb x0) er)))).
Proof.
    i.

    
    s in CA; rrdes CB; invc SVQ.
    dup CPa; rrdes CPa; clean.
    apply HP in CA; s in CA; des.
    clear HP ST.

    eapply pesim_stepb.
    i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
    { econs 1; inst_pairs; eapply Rok; eauto. }
    { inst_pairs; s; noemp; rr; esplits; eauto. }

    eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0 SV HV FIN FIN0.
    +
      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - readpc EXT0 Eladdr laddr.
          unfold pop; eauto.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
        readpc EXT0 Eladdr laddr.
      }
      { depmcplus1; eauto using alloc_valid. }
      s; setnewloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - readpc EXT0 Eladdr laddr.
        - rewrite 2!pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
          prep_store; eauto.
      }
      { s; tmext_simpl; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - econs; s; unfold mrf; tmext_simpl.
          eapply pmread_plus; eauto.
          symmetry; rewrite pmplus_swap LST SP.
          rewrite -!pmplus_assoc; eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - econs; s; unfold mrf; tmext_simpl.
          eapply pmread_plus; eauto.
          (rhs rewrite pmplus_swap); eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - readpc EXT0 Eladdr laddr.
        - rewrite pmcells_cons -pmplus_assoc.
          prep_store; eauto.
      }
      { s; tmext_simpl; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }
      fold newloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - econs; s; unfold mrf; tmext_simpl.
          eapply pmread_plus; first (exact CA); eauto.
          (rhs rewrite pmplus_swap); eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [by vauto|eapply Slpc].
        readpc EXT0 Eladdr laddr.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [by vauto|eapply Sopr].
        - readpc EXT0 Eladdr laddr.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [by vauto|eapply Sjz].
        - readpc EXT0 Eladdr laddr.
        - unfold mrf; tmext_simpl; auto.
      }
      { depmcplus0. }

      eapply dance_O.
      instantiate (1 := fun _ _ _ cf _ _ _ _ _ _ _ _ => depmach1 cf); s; eauto.

    + i; rrdes XTRb RLb; clarify.
      s in Ec; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { inst_pairs; rr; esplits; s; noemp; eauto. }
    
      i; frames_dep 3 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setcrf mrf'.
      destruct s' as [[[[]]]]; clean.
      setnewloc.
      set (s' := (mrf', P1, s5, s6, (fst s4, {newloc |=> [vax; r1 r_env]} ## snd s7))).

      assert (VLD': pmvalid ({newloc |=> [vax; r1 r_env]} ## snd s7)).
      { decomp_pmvalid HV; auto. }

      assert (s5 = s).
      { rrdes EXT0 EXT0 EXT0; ss; vauto. }
      subst s5.

      assert (EXT': ext (wlift mylw) s' (r1, P1, s, s6, s7)).
      { s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto.
        pmplus_eq'.
      }

      hexploit (@pmplus_from_read _ _ _ (snd s4) (snd s7)).
      { i; rrdes EXT0; apply EXT1; s; auto. }
      { split_pmvalid VLD'; eauto. }
      intro HH; des.

      exists s'; esplits; [by eauto|].

      rewrite <-subexp_tmext_l; last by auto.
      rewrite tmext_lift tgt_csplit src_csplit.
      eapply IHl.
      * eauto.
      * eauto.
      * eauto using ext_trans.
      * i; s in IN; des; vauto.
        apply RSV in IN; des.
        unfold s'; clean.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * eauto.
      * insts $ ((x, vax)::denva).
        i; apply list_assoc_cons_or in IN; des; subst.
        { rewrite list_assoc_cons_eq; esplits; eauto.
          tmext_simpl.
          eapply vclos_mon; eauto using ext_trans.
          i; eapply gk_mon; eauto using ext_trans.
        }
        rewrite tmext_neq; [|by vauto].
        rewrite list_assoc_cons_neq; [|by ii; subst x0; vauto].
        apply RDV in IN0; des.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * unfold s', mrf'; s; tmext_simpl; esplits
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
        rewrite -ENV0 in ENV; eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto.
        rewrite HH; pmplus_eq'.
      * s; rewrite DORD; eauto.
      * i; simpl map in IN.
        apply in_consD in IN; des; subst.
        { rewrite tmext_eq in FH; eauto using ts_gw_vclos_closed. }
        rewrite tmext_neq in FH; eauto.
      * rewrite <-tmext_lift; eapply weakclo_tmext; eauto.
        i; exfalso; eauto using ts_gw_vclos_closed.
      * vauto.
      * unfold s'; simpl_proj; econs; symmetry.
        rewrite HH Eladdr; simpl tr_exp.
        rewrite !pmcells_cons map_app pmcells_app.
        rewrite <-!pmplus_assoc.
        match goal with |- context[length ?x] => set (len := length x) end.
        
        instantiate (2 := 12 + length (map enc (tr_exp e n (map fst F) (map fst G))) + n).
        match goal with |- context[length ?x] => change (length x) with len end. 
        replace (_ - (12 + _ + n)) with (laddr - n); last omega.
        unfold len; s; rhs rewrite <-!(pmplus_swap {laddr - n |=> _}).
        rewrite !pmcells_cons map_app pmcells_app plus_comm !map_length; pmplus_eq'.
      * eauto.
      * 
        clear SV HV FIN.
        destruct s'0 as [[[[]]]]; clean.
        i; ts_prepare.

        hexploit (@pmplus_from_read _ _ _ ({newloc |=> [vax; r1 r_env]} ## snd s7) (snd s9)).
        { i; rrdes EXT0; apply EXT1; s; eauto. }
        { eauto. }
        intro HH'; des.

        assert (EXT'': ext (wlift mylw) (mrf, P2, s5, s8, s9) (r0, P0, s, s3, s4))
          by (repeat (eapply ext_trans; eauto)).

        eapply pesim_dance
        ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
        ; clear RLa0 SV HV FIN.

        { dance_init.
          { eapply Sok; [s; i; omega|eapply Slpc].
            readpc EXT'' Eladdr laddr.
          }
          { depmcplus0. }

          dance_init.
          { eapply Sok; [by vauto|eapply Sopr].
            - readpc EXT'' Eladdr laddr.
            - eauto.
            - eauto.
            - eauto.
          }
          { depmcplus0. }
          tmext_uniq.

          dance_init.
          { eapply Sok; [by vauto|eapply Sjmp].
            - readpc EXT'' Eladdr laddr.
            - eauto.
          }
          { depmcplus0. }

          dance_init.
          { eapply Sok; [by vauto|eapply Sld].
            - readpc EXT'' Eladdr laddr.
            - econs; s; unfold mrf; tmext_simpl.
              rewrite ENV1; s; unfold mrf'; tmext_simpl.
              eapply pmread_plus; eauto.
              rhs rewrite HH' !pmcells_cons -!pmplus_assoc -!(pmplus_swap {S newloc |-> _}).
              eauto.
          }
          { depmcplus0. }

          eapply dance_O.
          instantiate (1 := fun _ _ _ => depmach0); s; eauto.
        }

        
        i; rrdes XTRb RLb0; clarify.
        s in Ec; clarify.
        esplits; first 1 last.
        { right; split; eauto using wfo_le_n_add. }
        { inst_pairs; rr; esplits; s; noemp; eauto. }
    
        i; frames_nodep RLa cfas.
        destruct ea'; clear FIN.
        destruct_tgt_real RLa.
        simpl lang_tgt.chp in *.
        setcrf mrf''.
        set (s'' := (mrf'', P2, env, s8, s9)).

        assert (VLD'': pmvalid (snd s9)).
        { split_pmvalid HV; auto. }

        assert (EXT''': ext (wlift mylw) s'' (mrf, P2, env, s8, s9)).
        { unfold s'', mrf''; clear -HV VLD''.
          s; splits; tmext_simpl; auto.
        }

        exists s''; esplits; [by eauto|].

        hexploit (KSIM g'0); clear KSIM.
        { eauto. }
        { eauto. }
        { instantiate (1:=s'').
          repeat (eapply ext_trans; eauto).
        }
        { s; unfold mrf''; tmext_simpl; auto. }
        { s; unfold mrf''; tmext_simpl; rewrite SP0.
          s; unfold mrf'; tmext_simpl; rewrite SP; omega.
        }
        { s; unfold mrf''; tmext_simpl; rewrite RET0.
          s; unfold mrf'; tmext_simpl; rewrite RET; auto.
        }
        { s; unfold mrf''; tmext_simpl; rewrite LST0; s; auto. }
        { eapply vclos_mon; eauto.
          i; eapply gk_mon; eauto; repeat eapply ext_trans; eauto.
        }
        intro X; simpl tr_exp in X.
        exploit X; clear X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto]; s.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf''; cst := freest0; chp := hp0 |}).
            rrdes CPa; ss; subst; rr; esplits; s; unfold mrf''; tmext_simpl; eauto.
          - dup EXT'''; rrdes EXT'''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        f_equal.
        { rewrite !map_length !app_length; s; f_equal.
          rewrite !app_length; s; rewrite !app_length; s; omega.
        }
        { rewrite -pmplus_assoc; auto. }
      * unfold s'; rr; esplits.
        { rr; esplits; eauto.
          - s; rr; esplits; clean; eauto.
            eapply tgt_confpred_pop with (popped := [va]); eauto.
            + unfold mrf'; tmext_simpl; s; auto.
            + rewrite LST in SV; unfold mrf'; tmext_simpl.
              rewrite -pmcells_sng pmplus_comm; decomp_pmvalid SV; auto.
          - dup EXT'; rrdes EXT'; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        { rr; esplits; eauto. }
        { s; noemp; rewrite pmplus_swap; f_equal.
          - rewrite LST pmplus_swap -pmplus_assoc -pmcells_sng.
            unfold mrf'; tmext_simpl; auto.
          - rewrite -pmplus_assoc (pmplus_swap refha); auto.
        }
        { s; noemp; eauto. }
Qed.



Lemma tr_exp_correct_case: forall F G tyl tyr ty e x el er
    (IH: ts_psim F G (Tsum tyl tyr) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (IHl: ts_psim F ((x, tyl)::G) ty (fun n => map enc (tr_exp el n (map fst F) (x::map fst G))) el)
    (IHr: ts_psim F ((x, tyr)::G) ty (fun n => map enc (tr_exp er n (map fst F) (x::map fst G))) er),
  ts_psim F G ty (fun n => map enc (tr_exp (Ecase e x el er) n (map fst F) (map fst G))) (Ecase e x el er).
Proof.
  i; rdes IH; rdes IHl; rdes IHr; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  unfold ts_psim_aux in IH.
  eapply IH with (kb:= (SKcase x _ _)::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.
  i; ts_prepare.

  invc VR REL; last by (exfalso; invc VR).

  - 
    eapply tr_exp_correct_case_inl with (g := g); eauto.

  - 

    
    s in CA; rrdes CB; invc SVQ.
    dup CPa; rrdes CPa; clean.
    apply HP in CA; s in CA; des.
    clear HP ST.

    eapply pesim_stepb.
    i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
    { econs 1; inst_pairs; eapply Rok; eauto. }
    { inst_pairs; s; noemp; rr; esplits; eauto. }

    eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0 SV HV FIN FIN0.
    +
      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - readpc EXT0 Eladdr laddr.
          unfold pop; eauto.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
        readpc EXT0 Eladdr laddr.
      }
      { depmcplus1; eauto using alloc_valid. }
      s; setnewloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - readpc EXT0 Eladdr laddr.
        - rewrite 2!pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
          prep_store; eauto.
      }
      { s; tmext_simpl; pmcells_order; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - econs; s; unfold mrf; tmext_simpl.
          eapply pmread_plus; eauto.
          symmetry; rewrite pmplus_swap LST SP.
          rewrite -!pmplus_assoc; eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - econs; s; unfold mrf; tmext_simpl.
          eapply pmread_plus; eauto.
          (rhs rewrite pmplus_swap); eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Ssto].
        - readpc EXT0 Eladdr laddr.
        - rewrite pmcells_cons -pmplus_assoc.
          prep_store; eauto.
      }
      { s; tmext_simpl; pmcells_fold.
        depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
      }
      fold newloc.

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - econs; s; unfold mrf; tmext_simpl.
          eapply pmread_plus; first (exact CA0); eauto.
          (rhs rewrite pmplus_swap); eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [by vauto|eapply Slpc].
        readpc EXT0 Eladdr laddr.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { eapply Sok; [by vauto|eapply Sopr].
        - readpc EXT0 Eladdr laddr.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }
      tmext_uniq.

      dance_init.
      { destruct n0; [by vauto|].
        eapply Sok; [by vauto|eapply Sjnz].
        - readpc EXT0 Eladdr laddr.
        - unfold mrf; tmext_simpl; eauto.
        - eauto.
      }
      { depmcplus0. }

      eapply dance_O.
      instantiate (1 := fun _ _ _ cf _ _ _ _ _ _ _ _ => depmach1 cf); s; eauto.

    + i; rrdes XTRb RLb; clarify.
      s in Ec; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { inst_pairs; rr; esplits; s; noemp; eauto. }
    
      i; frames_dep 3 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setcrf mrf'.
      destruct s' as [[[[]]]]; clean.
      setnewloc.
      set (s' := (mrf', P1, s5, s6, (fst s4, {newloc |=> [vay; r1 r_env]} ## snd s7))).

      assert (VLD': pmvalid ({newloc |=> [vay; r1 r_env]} ## snd s7)).
      { decomp_pmvalid HV; auto. }

      assert (s5 = s).
      { rrdes EXT0 EXT0 EXT0; ss; vauto. }
      subst s5.

      assert (EXT': ext (wlift mylw) s' (r1, P1, s, s6, s7)).
      { s; splits; tmext_simpl; auto.
        i; eapply pmread_plus; eauto.
        pmplus_eq'.
      }

      hexploit (@pmplus_from_read _ _ _ (snd s4) (snd s7)).
      { i; rrdes EXT0; apply EXT1; s; auto. }
      { split_pmvalid VLD'; eauto. }
      intro HH; des.

      exists s'; esplits; [by eauto|].

      rewrite <-subexp_tmext_l; last by auto.
      rewrite tmext_lift tgt_csplit src_csplit.
      eapply IHr.
      * eauto.
      * eauto.
      * eauto using ext_trans.
      * i; s in IN; des; vauto.
        apply RSV in IN; des.
        unfold s'; clean.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * eauto.
      * insts $ ((x, vay)::denva).
        i; apply list_assoc_cons_or in IN; des; subst.
        { rewrite list_assoc_cons_eq; esplits; eauto.
          tmext_simpl.
          eapply vclos_mon; eauto using ext_trans.
          i; eapply gk_mon; eauto using ext_trans.
        }
        rewrite tmext_neq; [|by vauto].
        rewrite list_assoc_cons_neq; [|by ii; subst x0; vauto].
        apply RDV in IN0; des.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * unfold s', mrf'; s; tmext_simpl; esplits
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
        rewrite -ENV0 in ENV; eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto.
        rewrite HH; pmplus_eq'.
      * s; rewrite DORD; eauto.
      * i; simpl map in IN.
        apply in_consD in IN; des; subst.
        { rewrite tmext_eq in FH; eauto using ts_gw_vclos_closed. }
        rewrite tmext_neq in FH; eauto.
      * rewrite <-tmext_lift; eapply weakclo_tmext; eauto.
        i; exfalso; eauto using ts_gw_vclos_closed.
      * vauto.
      * unfold s'; simpl_proj; econs; symmetry.
        rewrite HH Eladdr; simpl tr_exp.
        rewrite !pmcells_cons map_app pmcells_app !map_length.
        rewrite <-!pmplus_assoc.
        match goal with |- context[length ?x + length ?y] =>
          instantiate (2 := 15 + length x + length y + n)
        end.
        replace (_ - (15 + _ + _ + n)) with (laddr - n); last omega.
        s; rhs rewrite <-!(pmplus_swap {laddr - n |=> _}).
        rewrite ?pmcells_cons ?map_app ?pmcells_app ?map_length; s.
        rewrite ?pmcells_cons ?map_app ?pmcells_app ?map_length.
        rewrite -(plus_comm 6) !plus_assoc -!(plus_swap 6); s.
        rewrite -(plus_swap laddr); pmplus_eq'.
      * eauto.
      * 
        clear SV HV FIN.
        destruct s'0 as [[[[]]]]; clean.
        i; ts_prepare.

        hexploit (@pmplus_from_read _ _ _ ({newloc |=> [vay; r1 r_env]} ## snd s7) (snd s9)).
        { i; rrdes EXT0; apply EXT1; s; eauto. }
        { eauto. }
        intro HH'; des.

        assert (EXT'': ext (wlift mylw) (mrf, P2, s5, s8, s9) (r0, P0, s, s3, s4))
          by (repeat (eapply ext_trans; eauto)).

        eapply pesim_dance
        ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
        ; clear RLa0 SV HV FIN.

        { dance_init.
          { eapply Sok; [by vauto|eapply Sld].
            - readpc EXT'' Eladdr laddr.
            - econs; s; unfold mrf; tmext_simpl.
              rewrite ENV1; s; unfold mrf'; tmext_simpl.
              eapply pmread_plus; eauto.
              rhs rewrite HH' !pmcells_cons -!pmplus_assoc -!(pmplus_swap {S newloc |-> _}).
              eauto.
          }
          { depmcplus0. }

          eapply dance_O.
          instantiate (1 := depmach0); s; eauto.
        }

        
        i; rrdes XTRb RLb0; clarify.
        s in Ec; clarify.
        esplits; first 1 last.
        { right; split; eauto using wfo_le_n_add. }
        { inst_pairs; rr; esplits; s; noemp; eauto. }
    
        i; frames_nodep RLa cfas.
        destruct ea'; clear FIN.
        destruct_tgt_real RLa.
        simpl lang_tgt.chp in *.
        setcrf mrf''.
        set (s'' := (mrf'', P2, env, s8, s9)).

        assert (VLD'': pmvalid (snd s9)).
        { split_pmvalid HV; auto. }

        assert (EXT''': ext (wlift mylw) s'' (mrf, P2, env, s8, s9)).
        { unfold s'', mrf''; clear -HV VLD''.
          s; splits; tmext_simpl; auto.
        }

        exists s''; esplits; [by eauto|].

        hexploit (KSIM g'0); clear KSIM.
        { eauto. }
        { eauto. }
        { instantiate (1:=s'').
          repeat (eapply ext_trans; eauto).
        }
        { s; unfold mrf''; tmext_simpl; auto. }
        { s; unfold mrf''; tmext_simpl; rewrite SP0.
          s; unfold mrf'; tmext_simpl; rewrite SP; omega.
        }
        { s; unfold mrf''; tmext_simpl; rewrite RET0.
          s; unfold mrf'; tmext_simpl; rewrite RET; auto.
        }
        { s; unfold mrf''; tmext_simpl; rewrite LST0; s; auto. }
        { eapply vclos_mon; eauto.
          i; eapply gk_mon; eauto; repeat eapply ext_trans; eauto.
        }
        intro X; simpl tr_exp in X.
        exploit X; clear X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto]; s.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf''; cst := freest0; chp := hp0 |}).
            rrdes CPa; ss; subst; rr; esplits; s; unfold mrf''; tmext_simpl; eauto.
          - dup EXT'''; rrdes EXT'''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        f_equal.
        { rewrite !map_length !app_length; s; f_equal.
          rewrite !app_length; s; rewrite !app_length; s; omega.
        }
        { rewrite -pmplus_assoc; auto. }
      * unfold s'; rr; esplits.
        { rr; esplits; eauto.
          - s; rr; esplits; clean; eauto.
            eapply tgt_confpred_pop with (popped := [va]); eauto.
            + unfold mrf'; tmext_simpl; s; auto.
            + rewrite LST in SV; unfold mrf'; tmext_simpl.
              rewrite -pmcells_sng pmplus_comm; decomp_pmvalid SV; auto.
          - dup EXT'; rrdes EXT'; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        { rr; esplits; eauto. }
        { s; noemp; rewrite pmplus_swap; f_equal.
          - rewrite LST pmplus_swap -pmplus_assoc -pmcells_sng.
            unfold mrf'; tmext_simpl; auto.
          - rewrite -pmplus_assoc (pmplus_swap refha); auto.
        }
        { s; noemp; eauto. }
Grab Existential Variables.
  exact i.
Qed.



Lemma tr_exp_correct_ifnz: forall F G ty e el er
    (IH: ts_psim F G Tnat (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (IHl: ts_psim F G ty (fun n => map enc (tr_exp el n (map fst F) (map fst G))) el)
    (IHr: ts_psim F G ty (fun n => map enc (tr_exp er n (map fst F) (map fst G))) er),
  ts_psim F G ty (fun n => map enc (tr_exp (Eifnz e el er) n (map fst F) (map fst G))) (Eifnz e el er).
Proof.
  i; rdes IH; rdes IHl; rdes IHr; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  unfold ts_psim_aux in IH.
  eapply IH with (kb:= (SKifnz _ _)::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.
  i; ts_prepare.

  invc VR REL; last by (exfalso; invc VR).
  s in CA; s in CB; rrdes CB; invc SVQ.
  destruct n0.

  - 

    eapply pesim_stepb.
    i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
    { econs 1; inst_pairs; eapply Rok; eauto. }
    { inst_pairs; s; noemp; rr; esplits; eauto. }

    eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0 SV HV FIN FIN0.
    +
      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - readpc EXT0 Eladdr laddr.
          unfold pop; eauto.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Slpc].
        readpc EXT0 Eladdr laddr.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - readpc EXT0 Eladdr laddr.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - econs; s; unfold mrf; tmext_simpl.
          eapply pmread_plus; eauto.
          symmetry; rewrite pmplus_swap LST SP.
          rewrite -!pmplus_assoc; eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Sjz].
        - readpc EXT0 Eladdr laddr.
        - unfold mrf; tmext_simpl; auto.
      }
      { depmcplus0. }

      eapply dance_O.
      instantiate (1 := fun _ _ _ _ => depmach0); s; eauto.

    + i; rrdes XTRb RLb; clarify.
      s in Ec; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { inst_pairs; rr; esplits; s; noemp; eauto. }
    
      i; frames_dep 3 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setcrf mrf'.
      destruct s' as [[[[]]]]; clean.
      setnewloc.
      set (s' := (mrf', P1, s5, s6, (fst s4, snd s7))).

      assert (VLD': pmvalid (snd s7)).
      { split_pmvalid HV; auto. }

      assert (s5 = s).
      { rrdes EXT0 EXT0 EXT0; ss; vauto. }
      subst s5.

      assert (EXT': ext (wlift mylw) s' (r1, P1, s, s6, s7)).
      { s; splits; tmext_simpl; auto. }

      hexploit (@pmplus_from_read _ _ _ (snd s4) (snd s7)).
      { i; rrdes EXT0; apply EXT1; s; auto. }
      { split_pmvalid VLD'; eauto. }
      intro HH; des.

      exists s'; esplits; [by eauto|].

      rewrite tgt_csplit src_csplit.
      eapply IHr.
      * eauto.
      * eauto.
      * eauto using ext_trans.
      * i; s in IN; des; vauto.
        apply RSV in IN; des.
        unfold s'; clean.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * eauto.
      * insts $ denva.
        i; apply RDV in IN; des.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * unfold s', mrf'; s; tmext_simpl; esplits
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
        rewrite -ENV0 in ENV; eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto.
      * s; rewrite DORD; eauto.
      * i; simpl map in IN; eauto.
      * auto.
      * vauto.
      * unfold s'; simpl_proj; econs; symmetry.
        rewrite HH Eladdr; simpl tr_exp.
        rewrite map_app pmcells_app.
        rewrite <-!pmplus_assoc.
        match goal with |- context[length ?x] => set (len := length x) end.
        match goal with |- context[length ?x] => change (length x) with len end. 
        instantiate (2 := 5 + length (map enc (tr_exp e n (map fst F) (map fst G))) + n).
        match goal with |- context[length ?x] => change (length x) with len end. 
        replace (_ - (5 + _ + n)) with (laddr - n); last omega.
        unfold len; s; rhs rewrite <-!(pmplus_swap {laddr - n |=> _}).
        rewrite !pmcells_cons map_app pmcells_app plus_comm !map_length; pmplus_eq'.
      * eauto.
      * 
        clear SV HV FIN.
        destruct s'0 as [[[[]]]]; clean.
        i; ts_prepare.

        hexploit (@pmplus_from_read _ _ _ (snd s7) (snd s9)).
        { i; rrdes EXT0; apply EXT1; s; eauto. }
        { eauto. }
        intro HH'; des.

        assert (EXT'': ext (wlift mylw) (mrf, P2, s5, s8, s9) (r0, P0, s, s3, s4))
          by (repeat (eapply ext_trans; eauto)).

        eapply pesim_dance
        ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
        ; clear RLa0 SV HV FIN.

        { dance_init.
          { eapply Sok; [s; i; omega|eapply Slpc].
            readpc EXT'' Eladdr laddr.
          }
          { depmcplus0. }

          dance_init.
          { eapply Sok; [by vauto|eapply Sopr].
            - readpc EXT'' Eladdr laddr.
            - eauto.
            - eauto.
            - eauto.
          }
          { depmcplus0. }
          tmext_uniq.

          dance_init.
          { eapply Sok; [by vauto|eapply Sjmp].
            - readpc EXT'' Eladdr laddr.
            - eauto.
          }
          { depmcplus0. }

          eapply dance_O.
          instantiate (1 := fun _ _ => depmach0); s; eauto.
        }

        
        i; rrdes XTRb RLb0; clarify.
        s in Ec; clarify.
        esplits; first 1 last.
        { right; split; eauto using wfo_le_n_add. }
        { inst_pairs; rr; esplits; s; noemp; eauto. }
    
        i; frames_nodep RLa cfas.
        destruct ea'; clear FIN.
        destruct_tgt_real RLa.
        simpl lang_tgt.chp in *.
        setcrf mrf''.
        set (s'' := (mrf'', P2, env, s8, s9)).

        assert (VLD'': pmvalid (snd s9)).
        { split_pmvalid HV; auto. }

        assert (EXT''': ext (wlift mylw) s'' (mrf, P2, env, s8, s9)).
        { unfold s'', mrf''; clear -HV VLD''.
          s; splits; tmext_simpl; auto.
        }

        exists s''; esplits; [by eauto|].

        hexploit (KSIM g'0); clear KSIM.
        { eauto. }
        { eauto. }
        { instantiate (1:=s'').
          repeat (eapply ext_trans; eauto).
        }
        { s; unfold mrf''; tmext_simpl; auto.
          rewrite ENV1; s; unfold mrf'; tmext_simpl; auto.
        }
        { s; unfold mrf''; tmext_simpl; rewrite SP0.
          s; unfold mrf'; tmext_simpl; rewrite SP; omega.
        }
        { s; unfold mrf''; tmext_simpl; rewrite RET0.
          s; unfold mrf'; tmext_simpl; rewrite RET; auto.
        }
        { s; unfold mrf''; tmext_simpl; rewrite LST0; s; auto. }
        { eapply vclos_mon; eauto.
          i; eapply gk_mon; eauto; repeat eapply ext_trans; eauto.
        }
        intro X; simpl tr_exp in X.
        exploit X; clear X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto]; s.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf''; cst := freest0; chp := hp0 |}).
            rrdes CPa0; ss; subst; rr; esplits; s; unfold mrf''; tmext_simpl; eauto.
          - dup EXT'''; rrdes EXT'''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        f_equal.
        { rewrite !map_length !app_length; s; f_equal.
          rewrite !app_length; s; omega.
        }
        { rewrite -pmplus_assoc; auto. }
      * unfold s'; rr; esplits.
        { rr; esplits; eauto.
          - s; rr; esplits; clean; eauto.
            eapply tgt_confpred_pop with (popped := [0]); eauto.
            + unfold mrf'; tmext_simpl; s; auto.
            + rewrite LST in SV; unfold mrf'; tmext_simpl.
              rewrite -pmcells_sng pmplus_comm; decomp_pmvalid SV; auto.
          - dup EXT'; rrdes EXT'; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        { rr; esplits; eauto. }
        { s; noemp; rewrite pmplus_swap; f_equal.
          - rewrite LST pmplus_swap -pmplus_assoc -pmcells_sng.
            unfold mrf'; tmext_simpl; auto.
          - rewrite -pmplus_assoc (pmplus_swap refha); auto.
        }
        { s; noemp; eauto. }

  - 

    eapply pesim_stepb.
    i; rrdes XTR RLb; clarify.
    rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
    { econs 1; inst_pairs; eapply Rok; eauto. }
    { inst_pairs; s; noemp; rr; esplits; eauto. }

    eapply pesim_dance
    ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
    ; clear RLa0 SV HV FIN FIN0.
    +
      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - readpc EXT0 Eladdr laddr.
          unfold pop; eauto.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Slpc].
        readpc EXT0 Eladdr laddr.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [s; i; omega|eapply Sopr].
        - readpc EXT0 Eladdr laddr.
        - eauto.
        - eauto.
        - eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Sld].
        - readpc EXT0 Eladdr laddr.
        - econs; s; unfold mrf; tmext_simpl.
          eapply pmread_plus; eauto.
          symmetry; rewrite pmplus_swap LST SP.
          rewrite -!pmplus_assoc; eauto.
      }
      { depmcplus0. }

      dance_init.
      { eapply Sok; [by vauto|eapply Sjnz].
        - readpc EXT0 Eladdr laddr.
        - unfold mrf; tmext_simpl; auto.
        - eauto.
      }
      { depmcplus0. }

      eapply dance_O.
      instantiate (1 := fun _ _ _ _ => depmach0); s; eauto.

    + i; rrdes XTRb RLb; clarify.
      s in Ec; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { inst_pairs; rr; esplits; s; noemp; eauto. }
    
      i; frames_dep 3 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setcrf mrf'.
      destruct s' as [[[[]]]]; clean.
      setnewloc.
      set (s' := (mrf', P1, s5, s6, (fst s4, snd s7))).

      assert (VLD': pmvalid (snd s7)).
      { split_pmvalid HV; auto. }

      assert (s5 = s).
      { rrdes EXT0 EXT0 EXT0; ss; vauto. }
      subst s5.

      assert (EXT': ext (wlift mylw) s' (r1, P1, s, s6, s7)).
      { s; splits; tmext_simpl; auto.  }


      hexploit (@pmplus_from_read _ _ _ (snd s4) (snd s7)).
      { i; rrdes EXT0; apply EXT1; s; auto. }
      { split_pmvalid VLD'; eauto. }
      intro HH; des.

      exists s'; esplits; [by eauto|].

      rewrite tgt_csplit src_csplit.
      eapply IHl.
      * eauto.
      * eauto.
      * eauto using ext_trans.
      * i; s in IN; des; vauto.
        apply RSV in IN; des.
        unfold s'; clean.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * eauto.
      * insts $ denva.
        i; apply RDV in IN; des.
        esplits; eauto.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * unfold s', mrf'; s; tmext_simpl; esplits
        ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
        rewrite -ENV0 in ENV; eapply reprlist_mon; eauto.
        i; eapply pmread_plus; eauto.
      * s; rewrite DORD; eauto.
      * i; simpl map in IN; eauto.
      * auto.
      * vauto.
      * unfold s'; simpl_proj; econs; symmetry.
        rewrite HH Eladdr; simpl tr_exp.
        rewrite map_app pmcells_app !map_length.
        rewrite <-!pmplus_assoc.
        match goal with |- context[length ?x + length ?y] =>
          instantiate (2 := 8 + length x + length y + n)
        end.
        replace (_ - (8 + _ + _ + n)) with (laddr - n); last omega.
        s; rhs rewrite <-!(pmplus_swap {laddr - n |=> _}).
        rewrite ?pmcells_cons ?map_app ?pmcells_app ?map_length; s.
        rewrite ?pmcells_cons ?map_app ?pmcells_app ?map_length.
        rewrite -(plus_comm 7) !plus_assoc -!(plus_swap 7); s.
        rewrite -(plus_swap laddr); pmplus_eq'.
      * eauto.
      * 
        clear SV HV FIN.
        destruct s'0 as [[[[]]]]; clean.
        i; ts_prepare.

        hexploit (KSIM g'0); clear KSIM.
        { eauto. }
        { eauto. }
        { repeat (eapply ext_trans; eauto). }
        { s; rewrite ENV1; s; unfold mrf'; tmext_simpl; auto. }
        { s; rewrite SP0.
          s; unfold mrf'; tmext_simpl; rewrite SP; omega.
        }
        { s; rewrite RET0.
          s; unfold mrf'; tmext_simpl; rewrite RET; auto.
        }
        { s; rewrite LST0; s; auto. }
        { eapply vclos_mon; eauto. }
        intro X; simpl tr_exp in X.
        exploit X; clear X.
        { rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          rr; splits; [|by rr; esplits; eauto]; s.
          instantiate (1 := {| cpc := Nothing; crf := Just mrf; cst := freest0; chp := hp0 |}).
          rrdes CPa; ss; subst; rr; esplits; s; eauto.
        }
        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        f_equal.
        { rewrite !map_length !app_length; s; f_equal.
          rewrite !app_length; s; omega.
        }
        { rewrite -pmplus_assoc; auto. }
      * unfold s'; rr; esplits.
        { rr; esplits; eauto.
          - s; rr; esplits; clean; eauto.
            eapply tgt_confpred_pop with (popped := [S n0]); eauto.
            + unfold mrf'; tmext_simpl; s; auto.
            + rewrite LST in SV; unfold mrf'; tmext_simpl.
              rewrite -pmcells_sng pmplus_comm; decomp_pmvalid SV; auto.
          - dup EXT'; rrdes EXT'; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        { rr; esplits; eauto. }
        { s; noemp; rewrite pmplus_swap; f_equal.
          - rewrite LST pmplus_swap -pmplus_assoc -pmcells_sng.
            unfold mrf'; tmext_simpl; auto.
          - rewrite -pmplus_assoc (pmplus_swap refha); auto.
        }
        { s; noemp; eauto. }
Grab Existential Variables.
  exact i.
Qed.
