(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis ssr.
Require Import lang_src_lsi lang_mid lang_spec.
Require Import model model_common model_properties gw.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


(** * Properties about [ms_gw]. *)

Lemma ms_gw_vclos_closed: forall (w: world_loca ms_gw) s r ty va vb
    (VC: vclos (wlift w) s r ty va vb)
    x (FV: fv_exp vb x),
  False.
Proof.
  i; invc VC REL; try rrdes CB; eauto.
  - rrdes VR; destruct vb; ss; subst; vauto.
  - invc VR; rrdes CB; eauto.
Qed.

Lemma ms_refconfrel_shape: forall g (s: state ms_gwsts) ha hb
    (RCR: ms_refconfrel g s ha hb),
  exists ha' hb',
    ha = (ha', Nothing) /\
    hb = (hb', Nothing, Nothing).
Proof.
  i; rrdes RCR; eauto.
Qed.


(** Convenience lemmas for reasoning about allocation. *)

Lemma confrel_alloc: forall g s s0 s1 s2 ty va vb ha hb la lb
    (SIM: vclos (wlift (lwemp ms_gw)) (s, s0, s1, s2) (g (s, s0, s1, s2)) ty va vb)
    (RH: confrel (wlift (lwemp ms_gw)) g (s, s0, s1, s2) (ha, Nothing) (hb, Just s0, Nothing))
    (VLDa: pmvalid ({la |-> va} ## ha))
    (VLDb: pmvalid ({lb |-> vb} ## hb))
    (GK: gknow (wlift (lwemp ms_gw)) g)
    (s1': refst) (S1: refst_alloc s1 ty la lb = s1'),
  confrel (wlift (lwemp ms_gw)) g (s, s0, s1', s2)
          ({la |-> va} ## ha, Nothing) ({lb |-> vb} ## hb, Just s0, Nothing).
Proof.
  i; rrdes RH; rr; rrdes WHY; clarify; noemp; subst.
  exists (({la |-> va} ## ha, Nothing) : conf mid_lsi). 
  ex @cemp.                                             
  exists (({lb |-> vb} ## hb, Just s0, Nothing) : conf src_lsi).
  ex @cemp.
  splits; noemp; rr; eauto.
  rrdes WHX; rr; clean.
  exists hax.
  exists ((({la |-> va}, Nothing) : conf mid_lsi) # hay).
  exists hbx.
  exists ((({lb |-> vb}, Nothing, Nothing) : conf src_lsi) # hby).
  splits; rrdes WHX0 WHY CPa CPb; noemp; clarify; noemp; subst
  ; s in HPB; noemp; clarify.
  - rr; esplits; s; eauto using pmvalid_plus_cell.
    + i; eapply pmread_plusD in ALC; des.
      * eapply pmread_cellD in ALC; des; subst.
        rewrite <-S1; unfold refst_alloc; eauto 10.
      * hexploit DOMa; eauto; i; des.
        rewrite <-S1; unfold refst_alloc; eauto.
    + i; eapply pmread_plusD in ALC; des.
      * eapply pmread_cellD in ALC; des; subst.
        rewrite <-S1; unfold refst_alloc; eauto 10.
      * hexploit DOMb; eauto; i; des.
        rewrite <-S1; unfold refst_alloc; eauto.
    + rewrite <-S1; unfold refst_alloc; i; des.
      * hexploit RELV; eauto; i; des.
        esplits; eauto using pmread_plus, pmplus_comm.
        eapply vclos_mon; eauto; [|i; eapply (gk_mon GK); eauto]
        ; rr; repeat split; eauto using refst_alloc_ext.
      * subst; esplits; s; eauto using pmread_plus, pmread_cell.
        eapply vclos_unlift in SIM.
        eapply vclos_mon; eauto; [|i; eapply (gk_mon GK); eauto]
        ; rr; repeat split; eauto using refst_alloc_ext.
  - s; noemp; clarify.
Qed.

Lemma confrel_newa: forall {lw: world_loca ms_gw} {g s0 s1 s2 s3 ha hb la va}
    (RH: confrel (wlift lw) g (s0, s1, s2, s3) (ha, Nothing) (hb, Just s1, Nothing))
    (VLD: pmvalid ({la |-> va} ## ha)),
  forall ty lb', ~ s2 ty la lb'.
Proof.
  i; rrdes RH WHX WHY0; clarify.
  specialize (RELV _ _ _ FH); des.
  destruct hax0, hay; ss; noemp; clarify.
  rrdes RDa; subst.
  rewrite <-!pmplus_assoc, (pmplus_swap h), pmplus_assoc in VLD.
  apply pmvalid_plusD in VLD; des.
  cut (la <> la); eauto; eapply pmvalid_cell_cell; eauto.
Qed.

Lemma confrel_newb: forall {lw: world_loca ms_gw} {g s0 s1 s2 s3 ha hb lb vb}
    (RH: confrel (wlift lw) g (s0, s1, s2, s3) (ha, Nothing) (hb, Just s1, Nothing))
    (VLD: pmvalid ({lb |-> vb} ## hb)),
  forall ty la', ~ s2 ty la' lb.
Proof.
  i; rrdes RH WHX WHY0; clarify.
  specialize (RELV _ _ _ FH); des.
  destruct hbx0 as [[]], hby as [[]]; ss; noemp; clarify.
  rrdes RDb; subst.
  rewrite <-!pmplus_assoc, (pmplus_swap h), pmplus_assoc in VLD.
  apply pmvalid_plusD in VLD; des.
  cut (lb <> lb); eauto; eapply pmvalid_cell_cell; eauto.
Qed.


(** Tactic for destructing some basic world-related assumptions that show up
    in many proofs.
*)

Ltac prepare :=
  apply pesim_acquire; intros ? ? HR
  ; apply pesim_real; intros ? _ RLa _
  ; let RLa' := fresh in assert (RLa':=RLa)
  ; destruct RLa' as (h & ee & Ec & Em & PV & FIN)
  ; let HR' := fresh in assert (HR':=HR)
  ; destruct HR' as (hax & hay & hbx & hby & WHX & WHY & HPA & HPB)
  ; destruct WHX as (hax0 & hay0 & hbx0 & hby0 & WHX0 & WHY0 & HPA0 & HPB0)
  ; destruct WHY as (WHY & WHY1)
  ; destruct WHX0 as (CPa & CPb); rrdes CPa CPb
  ; hexploit (ms_refconfrel_shape WHY0); intro XX; desH XX
  ; clarify; s in Ec; simpl cplus in HR; noemp; clarify.


Lemma stutter: forall i i' ea ty kim kis g si s cb cfb mb
    (RLb: real src_lsi (cb # cfb) mb)
    (SIM: pesim bote (wfo_nat (i' + i)) kim kis g si s false ty ea cb),
  exists j (eb' : conf src_lsi) (mb' : mach src_lsi),
     pesim (bote: esimtyp (wlift (lwemp ms_gw))) j kim kis g si s false ty ea eb' /\
     (tc (step src_lsi etau) mb mb' \/ mb = mb' /\ wfo_le_n 1 j (wfo_nat (1 + i' + i))) /\
     real src_lsi (eb' # cfb) mb'.
Proof.
  i; do 3 eexists; match goal with |- ?P /\ ?Q /\ ?R => cut (Q /\ R /\ P) end
  ; [by intuition|splits]; eauto.
  right; split; auto. 
  econs; eauto.
  eapply wfo_nat_lt_mon; omega.
Qed.

Lemma dolet: forall kim kis i g si s ty x v envm e es e' hm hs envs
    (RLm: mid_real (hm, Just (envm, IElet e' (inl x) e)) (hm, (envm, IElet e' (inl x) e)))
    (EVAL: eval envm e' = Some v)
    (SIM: esim (wlift (lwemp ms_gw)) (wfo_nat i) kim kis g si s false
               ty (hm, Just ((inl (inr x), v)::envm, e)) (hs, Just envs, Just es))
    (PV: pmvalid hm)
    (FIN: pmfinite hm),
   esim (wlift (lwemp ms_gw)) (wfo_nat (1 + i)) kim kis g si s false ty
        (hm, Just (envm, IElet e' (inl x) e))
        (hs, Just envs, Just es). 
Proof.
  i; eapply pesim_dtsteps; [by eauto| | |by eauto|].
  - eapply dtsteps_some; [by eapply mid_wdet; s; eauto| |].
    + instantiate (1:= (_,_)); econs; instantiate (2:= _::_).
      eapply IRlet; unfold eval; eauto.
    + eapply dtsteps_none.
  - rr; eauto 10.
  - simpl cplus at 2; noemp; i; eapply stutter; eauto.
Qed.
