(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw_common gw ssr util ts.
Require Import src2tgt src2tgt_correct_aux src2tgt_correct_refio.
Require Import src2tgt_correct_fun src2tgt_correct_atom src2tgt_correct_exmu.
Require Import src2tgt_correct_branch src2tgt_correct_rest.

Unset Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Theorem tr_mod_wf: forall FI FO m
    (WF: mhastype FI m FO),
  tgt_wfmod (tr_mod (map fst FI) m) (map fst FO).
Proof.
  unfold tr_mod, tgt_wfmod; s; i.
  set (n := (_ + _ + _)); generalize n; clear n.
  rr; splits.
  - destruct WF; clear FLI FLO.
    rewrite app_nil_r.
    generalize 0; ginduction MHT; vauto.
    s; i; f_equal.
    erewrite IHMHT; rewrite map_app; eauto.
  - 
    destruct WF.
    i; des; vauto; rhs rewrite <-plus_0_l.
    revert INl; generalize 0; ginduction MHT; [by vauto|].
    i; s in INl; des; clarify.
    + depdes v; destruct e_of; vauto; s in FLO; des; invc WTV HT; s; omega.
    + s in FLO; des.
      hexploit IHMHT; eauto.
      * apply is_thunkenv_app; s; eauto.
      * rewrite map_app; s; eauto.
      * i; s. rewrite app_length plus_assoc map_length.
        rewrite map_app in H; s; eauto.
Qed.



Lemma tr_exp_correct: forall e F G ty
    (WF: hastype F G e ty),
  ts_psim F G ty (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e.
Proof.
  i; invc WF; clear LCE0 LCT LCE; ginduction HT; i.
  - apply tr_exp_correct_input.
  - apply tr_exp_correct_output; auto.
  - apply tr_exp_correct_var; auto.
  - apply tr_exp_correct_lbl; auto.
  - apply tr_exp_correct_unit.
  - apply tr_exp_correct_nat.
  - apply tr_exp_correct_bop; auto.
  - apply tr_exp_correct_ifnz; auto.
  - eapply tr_exp_correct_pair; auto.
  - eapply tr_exp_correct_fst; eauto.
  - eapply tr_exp_correct_snd; eauto.
  - eapply tr_exp_correct_inl; eauto.
  - eapply tr_exp_correct_inr; eauto.
  - eapply tr_exp_correct_case; eauto.
  - eapply tr_exp_correct_roll; auto.
  - eapply tr_exp_correct_unroll; auto.
  - eapply tr_exp_correct_fix; eauto.
  - eapply tr_exp_correct_app; eauto.
  - eapply tr_exp_correct_gen; eauto.
  - eapply tr_exp_correct_inst; eauto.
  - eapply tr_exp_correct_pack; eauto.
  - eapply tr_exp_correct_unpack; ss; eauto.
  - eapply tr_exp_correct_new; auto.
  - eapply tr_exp_correct_read; auto.
  - eapply tr_exp_correct_write; eauto.
  - eapply tr_exp_correct_refeq; eauto.
Qed.



Lemma tr_toplvl_correct_fix: forall FI FI' ancha suba subb s rest n va (vb: val) (ty: flextyp) envf ty1 ty2
    (EFI: FI' = map_pair id flext_of FI)
    (SUBa: map fst suba = map fst FI)
    (SUBb: map fst subb = map fst FI)
    (Ety: Tarrow ty1 ty2 = ty)
    (WF: hastype FI' nil vb ty)
    (FUN: ts_db s (ts_fun (ancha + n)) va : Prop)
    (COD: ts_lhp s = {ancha |=> map snd suba} ##
                     {ancha + n |=> map enc (tr_toplvl vb n (map fst FI))} ## rest)
    (ENV: ts_env s = subb ++ envf)
    (VLD: pmvalid (ts_lhp s))
    (NZ: n <> 0),
  @tlsim _ _ ts_gw mylw FI suba subb s va vb ty.
Proof.
  ii; invc WF; destruct vb; destruct e_of; vauto
  ; inv HT; vauto; rewrite -?H0 -?H1 -?H2 in Ety; vauto.
  rewrite -Ety in H4; clarify.
  clear LCE0.
  assert (Efty: ty = FTarrow ty1 ty2).
  { apply flext_of_inj; destruct ty eqn:E; clarify. }
  rewrite Efty.

  hexploit tr_exp_correct; [econs|]; eauto.
  { rewrite Efty in LCT. s in LCT.
    ii; s in IN; des; subst; clarify.
  }
  { rewrite Efty in LCT; s in LCT; des; auto. }
  intro IH; simpl flext_of in IH; simpl lang_src.e_of in IH.
  rdes IH.

  eapply vclf_fix.
  { econs; apply LE; eauto. }
  { rr; splits; [by s; eauto|].
    ii; des; eapply hastype_fv in FH; eauto; vauto.
  }

  eapply gk_lsim; eauto.
  evar (i0: wfo_set (wlift mylw)).
  econs; [econs|].
  { econs; apply LE; eauto. }
  { rr; splits; [by s; eauto|].
    ii; des; eapply hastype_fv in FH; eauto; vauto.
  }

  insts $ i0.
  pcofix CIH; unfold i0 in *; clear i0.
  i; invc STK; [|by invc INSTb RF CB VR].
  eapply pesim_gknow; i; eapply pesim_gknow_; i.
  invc APPa APPb; invc RF CB VR; rrdes CA; s in VAL; clarify.
  destruct s' as [[[[]]]]; clean.
  ren PC; destruct s3.
  apply pesim_acquire; ts_prepare; setnewloc.

  clear CA CPb.

  hexploit (@pmplus_from_read _ _ _ (snd (snd s)) h).
  { i; rrdes EXT; apply EXT0; s.
    rrdes LE LE1; apply LE1; eauto.
  }
  { split_pmvalid VLD; eauto. }
  intro COD'; des; subst h.

  
  rewrite pmplus_swap in COD.
  assert (Epc: mpc = ancha + n).
  {
    hexploit tgt_vquery_fun_uniq; [by eauto| |by eauto|].
    - apply EXT, LE; eauto.
    - auto.
  }
  subst mpc.
  
  

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc, <-pmcells_sng in *; clear NST.

  eapply pesim_dance; first (by eauto); clear RLa0.
  + clear SV HV FIN.
    s in COD.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc' (ancha + n) COD.
        unfold push; eauto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - readpc' (ancha + n) COD.
      - unfold stack_top; tmext_simpl; prep_store; eauto.
    }
    { s; tmext_simpl.
      repeat match goal with |- context[?a + 1 - 1] =>
                             replace (a + 1 - 1) with a by omega
             end.
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc' (ancha + n) COD.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc]; [|by eauto].
      readpc' (ancha + n) COD.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' (ancha + n) COD.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { fold newloc; tmext_simpl; s; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc' (ancha + n) COD.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc]; [|by eauto].
      readpc' (ancha + n) COD.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.
    tmext_uniq.

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' (ancha + n) COD.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { fold newloc newloc0; tmext_simpl; s; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' (ancha + n) COD.
      - rewrite 2!pmcells_cons -!pmplus_assoc (pmplus_swap {newloc0 |-> _}).
        prep_store; eauto.
    }
    { tmext_simpl; s; pmcells_order; pmcells_fold; noemp.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    eapply dance_O; instantiate (1 := fun _ _ _ cf _ _ cf' _ _ => depmach2 cf cf').
    s; tmext_simpl; eauto.

  + i; rrdes XTRb RLb; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { eauto. }
    
    i; frames_dep 3 RLa XTRa cfas; frames_dep 2 RLa XTRa cfas; frames_nodep RLa cfas.
    destruct ea'; clear SV HV FIN.
    destruct_tgt_real RLa.
    simpl lang_tgt.chp in *.
    setnewloc; setcrf mrf'.

    assert (env = subb ++ envf) by
        (by rrdes EXT EXT EXT EXT2; rrdes LE LE LE LE3; rewrite -ENV -LE3).
    subst env.

    set (s' := (mrf', P, subb++envf, s2, ({mrf r_sp |-> mrf r_env} ## s3,
        {newloc |=> [mrf r_arg; newloc0]} ## {newloc0 |=> [mrf r_clo; 0]} ## snd (snd s) ## h''))).

    assert (VLD'': pmvalid ({newloc |=> [mrf r_arg; newloc0]} ##
                            {newloc0 |=> [mrf r_clo; 0]} ## snd (snd s) ## h'')).
    { clear -HV.
      rewrite -!(pmplus_swap hp) -!(pmplus_swap refha) in HV.
      eapply pmvalid_plusD, pmvalid_plusD; eauto.
    }

    assert (EXT'': ext (wlift mylw) s' (mrf, P, subb++envf, s2, (s3, snd (snd s) ## h''))).
    { clear -VLD''.
      s; splits; tmext_simpl; auto.
      i; eapply pmread_plus; eauto.
      pmplus_eq'.
    }

    exists s'; esplits; [exact EXT''|].
    eapply paco10_mon.
    rewrite tgt_csplit src_csplit !(tmext_lift e_of).
    eapply IH.
    * insts $ tid; auto.
    * eauto.
    * exact EXT''.
    * i; s in IN; des; vauto.
      assert (exists fty, lookup l FI = Some fty /\ ty'0 = fty).
      { apply list_assoc_map_pairD; auto. }
      des.
      assert (exists va', lookup l suba = Some va').
      { apply eq_sym in SUBa; eapply list_assoc_eqdom; eauto. }
      des.
      assert (exists vb', lookup l subb = Some vb').
      { apply eq_sym in SUBb; eapply list_assoc_eqdom; eauto. }
      des.
      eapply RS in H2; eauto.
      esplits; eauto.
      { s; rewrite list_assoc_app H5; eauto. }
      rewrite tsubst_name_tid. 
      eapply @vclgen_thunk in H2; apply vclos_intro in H2. 
      rewrite H3.
      eapply vclos_mon; eauto using ext_trans.
      i; eapply gk_mon; eauto using ext_trans.
    * rewrite SUBa map_fst_map_pair map_map; auto.
    * instantiate (1 := (x0, mrf r_arg)::(f0, mrf r_clo)::nil).
      i; apply list_assoc_cons_or in IN; des.
      { subst x; subst ty'0.
        rewrite list_assoc_cons_eq; esplits; eauto.
        tmext_simpl.
        rewrite tsubst_name_tid. 
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      }
      rewrite tmext_neq; [|by vauto].
      rewrite list_assoc_cons_neq; [|by ii; subst x; vauto].
      apply list_assoc_cons_or in IN0; des.
      { subst x ty'0.
        rewrite list_assoc_cons_eq; esplits; eauto.
        tmext_simpl.
        
        econs; apply vclgen_thunk, vclf_fix.
        { unfold s'; econs; rr; eauto. }
        { rr; splits; s; eauto. }
        eapply gk_lsim; eauto.
        econs; [econs|].
        { unfold s'; econs; rr; eauto. }
        { rr; splits; s; eauto. }
        i; eapply gres_incl_id.
        right; eapply CIH; eauto using ext_trans.
        fold tsubst_name in STK; rewrite !tsubst_name_tid in STK.
        set (xx := mkval _ _).
        assert (Exx: xx = Vfix f0 x0 e).
        { destruct xx eqn:Exx'.
          destruct (Vfix _ _ _) eqn:Eff.
          ss; clarify; unfold Vfix in Eff; clarify.
          f_equal; eauto using proof_irrelevance.
        }                     
        rewrite Exx.
        eauto.
      }
      rewrite tmext_neq; [|by vauto].
      rewrite list_assoc_cons_neq; [|by ii; vauto].
      vauto.
    * unfold s', mrf'; s; tmext_simpl; esplits
      ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
      auto.
    * auto.
    * i; simpl map in IN.
      apply in_consD in IN; des.
      { subst x; rewrite tmext_eq in FH; eauto using ts_gw_vclos_closed. }
      rewrite tmext_neq in FH; auto.
      apply in_consD in IN; des.
      { subst x; rewrite tmext_eq in FH; eauto using ts_gw_vclos_closed. }
      rewrite tmext_neq in FH; eauto.
    * rewrite <-tmext_lift; eapply weakclo_tmext; last first.
      { i; exfalso; eauto using ts_gw_vclos_closed. }
      rewrite <-tmext_lift; eapply weakclo_tmext; last first.
      { i; exfalso; eauto using ts_gw_vclos_closed. }
      eauto.
    * vauto.
    * unfold s'; simpl_proj; econs; symmetry.
      rewrite COD; simpl tr_exp.
      rewrite !pmcells_cons map_app pmcells_app.
      rewrite <-!pmplus_assoc.
      instantiate (2 := 9 + n).
      replace (_ - (9 + n)) with ancha; [|omega].
      rhs rewrite <-!(pmplus_swap {ancha |=> _}).
      rewrite map_fst_map_pair map_map.
      pmplus_eq'.
    * eauto.
    * 
      clear SV HV FIN FIN0.
      destruct s'0 as [[[[]]]]; clean.
      i; ts_prepare. 

      hexploit (@pmplus_from_read _ _ _ (snd (snd s)) (snd s5)).
      { i; rrdes EXT0; apply EXT1; s.
        eapply pmread_plus; eauto; pmplus_eq'.
      }
      { split_pmvalid VLD''; eauto. }
      intro EE; des; rewrite ->EE in *.

      eapply pesim_dance
      ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
      ; clear RLa0 SV HV FIN.

      { s in COD.
        rewrite map_fst_map_pair map_map.

        dance_init.
        { eapply Sok; [s; i; omega|eapply Sld].
          - readpc' (ancha + n) COD.
          - econs; s.
            eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [by vauto|eapply Sopr].
          - readpc' (ancha + n) COD.
            unfold pop; eauto.
          - eauto.
          - eauto.
          - eauto.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [s; i; omega|eapply Sld].
          - readpc' (ancha + n) COD.
          - econs; s; tmext_simpl.
            eapply pmread_plus; eauto.
            rewrite LST SP; unfold s'; clean.
            unfold mrf'; tmext_simpl.
            repeat match goal with |- context[?a + 1 + 1 - 2] =>
                                   replace (a + 1 + 1 - 2) with a by omega
                   end.                
            pmplus_eq'.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [by vauto|eapply Sjmp].
          - readpc' (ancha + n) COD.
          - eauto.
        }
        { depmcplus0. }

        eapply dance_O; instantiate (1 := fun _ _ _ => depmach0).
        s; tmext_simpl; eauto.
      }

      clear RLb0.
      i; rrdes XTRb RLb0; clarify.
      s in Ec; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { inst_pairs; rr; esplits; s; noemp; eauto. }
      
      i; frames_nodep RLa cfas.
      destruct ea'; clear FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setcrf mrf''.
      set (s'' := (mrf'', P0, env, s4, (s3, snd s5))).

      assert (VLD''': pmvalid (snd s5)).
      { rewrite EE; decomp_pmvalid HV; auto. }

      assert (EXT''': ext (wlift mylw) s'' (mrf0, P0, env, s4, s5)).
      { unfold s'', mrf''; clear -HV VLD'''.
        s; splits; tmext_simpl; auto.
      }

      exists s''; esplits; [by eauto|].
      rewrite tgt_csplit src_csplit.
      eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s''; rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; clean; eauto.
            eapply tgt_confpred_pop; eauto; last first.
            * rewrite LST SP in SV; unfold s' in SV; unfold mrf' in SV.
              s in SV; tmext_simplH SV.
              repeat match goal with H: context[?a + 1 - 1] |- _ =>
                                     replace (a + 1 - 1) with a in H by omega
                     end.
              unfold mrf''; rewrite pmplus_comm; tmext_simpl.
              rewrite SP; s; unfold mrf'; tmext_simpl.
              repeat match goal with |- context[?a + 1 + 1 - 2] =>
                                     replace (a + 1 + 1 - 2) with a by omega
                     end.
              rewrite (pmplus_swap {_|->_}) in SV.
              rewrite plus_comm in SV; simpl plus in SV.
              rewrite (pmcells_sng va) in SV; pmcells_foldH SV.
              
              decomp_pmvalid SV; eauto.
            * unfold mrf''; tmext_simpl; s; auto.
          + dup EXT'''; rrdes EXT'''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            * s; i; eapply gk_mon; eauto; split; eauto.
            * s; i; eapply gk_mon; eauto; split; eauto.
        - rr; esplits; eauto.
        - s; rewrite LST SP EE -!pmplus_assoc; noemp; s; f_equal.
          unfold mrf''; tmext_simpl; rewrite SP; s.
          unfold mrf'; tmext_simpl.
          rewrite -!(pmplus_swap freest).
          repeat match goal with |- context[?a + 1 + 1 - 2] =>
                                 replace (a + 1 + 1 - 2) with a by omega
                 end.
          repeat match goal with |- context[?a + 1 - 1] =>
                                 replace (a + 1 - 1) with a by omega
                 end.
          
          rewrite plus_comm (pmplus_swap {_ |-> _}); auto.
        - s; noemp; eauto.
      }

      invc RK; fold tsubst_name.
      rewrite RET; s; tmext_simpl; unfold mrf'; tmext_simpl.
      eapply ts_pesim_return.
      { s; unfold mrf''; tmext_simpl.
        rewrite tsubst_name_tid in VR.
        eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
      }
      { split; split; s.
        - rrdes EXT0 EXT0 EXT0.
          splits; auto.
          + unfold mrf''; tmext_simpl.
            rewrite SP; s; unfold mrf'; tmext_simpl.
            omega.
          + unfold mrf''; tmext_simpl; auto.
        - rrdes EXT0 EXT0. apply EXT2.
        - rrdes EXT0 EXT0. s in EXT1.
          rrdes EXT'' EXT''. s in EXT''0.
          unnw. eauto.
        - auto.
      }
    * unfold s'; rr; esplits.
      { rr; esplits; eauto.
        - s; rr; esplits; clean; eauto; [|by rr; eauto].
          rewrite pmcells_sng in CPa.
          eapply tgt_confpred_push; eauto.
          unfold mrf'; tmext_simpl; s; auto.
        - dup EXT''; rrdes EXT''; simpl_proj.
          eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
      }
      { rr; esplits; eauto. }
      { s. noemp; rewrite pmplus_swap; f_equal.
        rewrite 2!(pmplus_swap (hp ## _)) -!pmplus_assoc; eauto.
      }
      { s; noemp; eauto. }
    * i; vauto.
Grab Existential Variables.
  exact i.
Qed.



Lemma tr_toplvl_correct_gen: forall FI FI' ancha suba subb s rest n va (vb: val) (ty: flextyp) envf ty1
    (EFI: FI' = map_pair id flext_of FI)
    (SUBa: map fst suba = map fst FI)
    (SUBb: map fst subb = map fst FI)
    (Ety: Tforall ty1 = ty)
    (WF: hastype FI' nil vb ty)
    (FUN: ts_db s (ts_fun (ancha + n)) va : Prop)
    (COD: ts_lhp s = {ancha |=> map snd suba} ##
                     {ancha + n |=> map enc (tr_toplvl vb n (map fst FI))} ## rest)
    (ENV: ts_env s = subb ++ envf)
    (VLD: pmvalid (ts_lhp s))
    (NZ: n <> 0),
  @tlsim _ _ ts_gw mylw FI suba subb s va vb ty.
Proof.
  ii; invc WF; destruct vb; destruct e_of; vauto
  ; inv HT; vauto; rewrite -?H0 -?H1 -?H2 in Ety; vauto.
  rewrite -Ety in H4; clarify.
  clear LCE0.
  assert (Efty: ty = FTforall ty0).
  { apply flext_of_inj; destruct ty eqn:E; clarify. }
  rewrite Efty.

  hexploit tr_exp_correct; [econs|]; eauto.
  { rewrite Efty in LCT. s in LCT.
    ii; s in IN; des; subst; clarify.
  }
  { rewrite Efty in LCT; eapply tsubbvar_lctyp; eauto. }
  intro IH; simpl flext_of in IH; simpl lang_src.e_of in IH.
  clear H0; rdes IH.

  eapply vclf_gen.
  { econs; apply LE; eauto. }
  { rr; splits; [by s; eauto|].
    ii; des; eapply hastype_fv in FH; eauto; vauto.
  }

  eapply gk_lsim; eauto.
  evar (i0: wfo_set (wlift mylw)).
  econs; [econs|].
  { econs; apply LE; eauto. }
  { rr; splits; [by s; eauto|].
    ii; des; eapply hastype_fv in FH; eauto; vauto.
  }

  insts $ i0.
  pcofix CIH; unfold i0 in *; clear i0.
  i; invc STK; [by invc APPa RF VR|].
  eapply pesim_gknow; i; eapply pesim_gknow_; i.
  invc INSTa INSTb; invc RF RK VR; s in VAL; clarify.
  destruct s' as [[[[]]]]; clean.
  ren PC; destruct s3.
  apply pesim_acquire; ts_prepare; setnewloc.

  clear CA CPb.

  hexploit (@pmplus_from_read _ _ _ (snd (snd s)) h).
  { i; rrdes EXT; apply EXT0; s.
    rrdes LE LE1; apply LE1; eauto.
  }
  { split_pmvalid VLD; eauto. }
  intro COD'; des; subst h.

  
  rewrite pmplus_swap in COD.
  assert (Epc: mpc = ancha + n).
  {
    hexploit tgt_vquery_fun_uniq; [by eauto| |by eauto|].
    - apply EXT, LE; eauto.
    - auto.
  }
  subst mpc.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc, <-pmcells_sng in *; clear NST.

  eapply pesim_dance; first (by eauto); clear RLa0.
  - clear SV HV FIN.
    s in COD.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc' (ancha + n) COD.
        unfold push; eauto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Ssto].
      - readpc' (ancha + n) COD.
      - unfold stack_top; tmext_simpl; prep_store; eauto.
    }
    { s; tmext_simpl.
      repeat match goal with |- context[?a + 1 - 1] =>
                             replace (a + 1 - 1) with a by omega
             end.
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    eapply dance_O; instantiate (1 := fun _ => depmach0).
    s; tmext_simpl; eauto.

  - i; rrdes XTRb RLb; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { eauto. }
    
    i; frames_nodep RLa cfas.
    destruct ea'; clear SV HV FIN.
    destruct_tgt_real RLa.
    simpl lang_tgt.chp in *.
    setnewloc; setcrf mrf'.

    assert (env = subb ++ envf) by
        (by rrdes EXT EXT EXT EXT2; rrdes LE LE LE LE3; rewrite -ENV -LE3).
    subst env.

    set (s' := (mrf', P, subb++envf, s2, ({mrf r_sp |-> mrf r_env} ## s3, snd (snd s) ## h''))).

    assert (VLD'': pmvalid (snd (snd s) ## h'')).
    { clear -HV.
      eapply pmvalid_plusD, pmvalid_plusD; eauto.
    }

    assert (EXT'': ext (wlift mylw) s' (mrf, P, subb++envf, s2, (s3, snd (snd s) ## h''))).
    { clear -VLD''.
      s; splits; tmext_simpl; auto.
    }

    exists s'; esplits; [exact EXT''|].
    eapply paco10_mon.
    rewrite tgt_csplit src_csplit -(subexp_vid e).
    eapply (IH ([tyi//aa]>>tid)).
    + ii; destruct (LEM (x = aa)); subst.
      * rewrite tmext_eq; eauto.
      * rewrite tmext_neq; eauto.
    + eauto.
    + exact EXT''.
    + i; s in IN; des; vauto.
      assert (exists fty, lookup l FI = Some fty /\ ty' = fty).
      { apply list_assoc_map_pairD; auto. }
      des.
      assert (exists va', lookup l suba = Some va').
      { apply eq_sym in SUBa; eapply list_assoc_eqdom; eauto. }
      des.
      assert (exists vb', lookup l subb = Some vb').
      { apply eq_sym in SUBb; eapply list_assoc_eqdom; eauto. }
      des.
      eapply RS in H; eauto.
      esplits; eauto.
      { s; rewrite list_assoc_app H5; eauto. }
      erewrite (tsubname_agree _ _ tid).
      * rewrite tsubst_name_tid H0.
        eapply @vclgen_thunk in H; apply vclos_intro in H.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * i; clear - H0 H3 IN IN0.
        assert (x <> aa).
        { ii; subst; induction FI; vauto.
          destruct a; s in IN; des_ifs; subst.
          - apply H3; s; rewrite H0; eauto.
          - eapply IHFI; ss; i; des; eauto.
        }
        rewrite tmext_neq; eauto. 
    + rewrite SUBa map_fst_map_pair map_map; auto.
    + instantiate (1 := nil); vauto.
    + s; vauto.
    + auto.
    + i; vauto.
    + eauto.
    + vauto.
    + unfold s'; simpl_proj; econs; symmetry.
      rewrite COD; simpl tr_exp.
      rewrite !pmcells_cons map_app pmcells_app.
      rewrite <-!pmplus_assoc.
      instantiate (2 := 2 + n).
      replace (_ - (2 + n)) with ancha; [|omega].
      rhs rewrite <-!(pmplus_swap {ancha |=> _}).
      rewrite map_fst_map_pair map_map.
      pmplus_eq'.
    + eauto.
    + 
      clear SV HV FIN FIN0.
      destruct s'0 as [[[[]]]]; clean.
      i; ts_prepare.

      hexploit (@pmplus_from_read _ _ _ (snd (snd s)) (snd s5)).
      { i; rrdes EXT0; apply EXT1; s.
        eapply pmread_plus; eauto; pmplus_eq'.
      }
      { split_pmvalid VLD''; eauto. }
      intro EE; des; rewrite ->EE in *.

      eapply pesim_dance
      ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
      ; clear RLa0 SV HV FIN.

      { s in COD.
        rewrite map_fst_map_pair map_map.

        dance_init.
        { eapply Sok; [s; i; omega|eapply Sld].
          - readpc' (ancha + n) COD.
          - econs; s.
            eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [by vauto|eapply Sopr].
          - readpc' (ancha + n) COD.
            unfold pop; eauto.
          - eauto.
          - eauto.
          - eauto.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [s; i; omega|eapply Sld].
          - readpc' (ancha + n) COD.
          - econs; s; tmext_simpl.
            eapply pmread_plus; eauto.
            rewrite LST SP; unfold s'; clean.
            unfold mrf'; tmext_simpl.
            repeat match goal with |- context[?a + 1 + 1 - 2] =>
                                   replace (a + 1 + 1 - 2) with a by omega
                   end.                
            pmplus_eq'.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [by vauto|eapply Sjmp].
          - readpc' (ancha + n) COD.
          - eauto.
        }
        { depmcplus0. }

        eapply dance_O; instantiate (1 := fun _ _ _ => depmach0).
        s; tmext_simpl; eauto.
      }

      clear RLb0.
      i; rrdes XTRb RLb0; clarify.
      s in Ec; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { inst_pairs; rr; esplits; s; noemp; eauto. }
      
      i; frames_nodep RLa cfas.
      destruct ea'; clear FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setcrf mrf''.
      set (s'' := (mrf'', P0, env, s4, (s3, snd s5))).

      assert (VLD''': pmvalid (snd s5)).
      { rewrite EE; decomp_pmvalid HV; auto. }

      assert (EXT''': ext (wlift mylw) s'' (mrf0, P0, env, s4, s5)).
      { unfold s'', mrf''; clear -HV VLD'''.
        s; splits; tmext_simpl; auto.
      }

      exists s''; esplits; [by eauto|].
      rewrite tgt_csplit src_csplit.
      eapply pesim_release; last first; try by rr; s; eauto.
      { unfold s''; rr; esplits.
        - rr; esplits; eauto.
          + s; rr; esplits; clean; eauto.
            eapply tgt_confpred_pop; eauto; last first.
            * rewrite LST SP in SV; unfold s' in SV; unfold mrf' in SV.
              s in SV; tmext_simplH SV.
              repeat match goal with H: context[?a + 1 - 1] |- _ =>
                                     replace (a + 1 - 1) with a in H by omega
                     end.
              unfold mrf''; rewrite pmplus_comm; tmext_simpl.
              rewrite SP; s; unfold mrf'; tmext_simpl.
              repeat match goal with |- context[?a + 1 + 1 - 2] =>
                                     replace (a + 1 + 1 - 2) with a by omega
                     end.
              rewrite (pmplus_swap {_|->_}) in SV.
              rewrite plus_comm in SV; simpl plus in SV.
              rewrite (pmcells_sng va) in SV; pmcells_foldH SV.
              
              decomp_pmvalid SV; eauto.
            * unfold mrf''; tmext_simpl; s; auto.
          + dup EXT'''; rrdes EXT'''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            * s; i; eapply gk_mon; eauto; split; eauto.
            * s; i; eapply gk_mon; eauto; split; eauto.
        - rr; esplits; eauto.
        - s; rewrite LST SP EE -!pmplus_assoc; noemp; s; f_equal.
          unfold mrf''; tmext_simpl; rewrite SP; s.
          unfold mrf'; tmext_simpl.
          rewrite -!(pmplus_swap freest).
          repeat match goal with |- context[?a + 1 + 1 - 2] =>
                                 replace (a + 1 + 1 - 2) with a by omega
                 end.
          repeat match goal with |- context[?a + 1 - 1] =>
                                 replace (a + 1 - 1) with a by omega
                 end.
          
          rewrite plus_comm (pmplus_swap {_ |-> _}); auto.
        - s; noemp; eauto.
      }

      fold tsubst_name.
      rewrite RET; s; tmext_simpl; unfold mrf'; tmext_simpl.
      eapply ts_pesim_return.
      { s; unfold mrf''; tmext_simpl.
        rewrite tsub_name_bvar in VR.
        - s in VR; rewrite tmext_eq in VR.
          erewrite (tsubname_agree _ _ tid) in VR.
          + rewrite tsubst_name_tid in VR.
            eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
          + i; clear - LCT H3 IN.
            assert (x <> aa) by (ii; subst; vauto; s in H3; des; eauto).
            rewrite tmext_neq; eauto.
        - ii; destruct (LEM (x = aa)); subst.
          + rewrite tmext_eq; eauto.
          + rewrite tmext_neq; eauto. 
      }
      { split; split; s.
        - rrdes EXT0 EXT0 EXT0.
          splits; auto.
          + unfold mrf''; tmext_simpl.
            rewrite SP; s; unfold mrf'; tmext_simpl.
            omega.
          + unfold mrf''; tmext_simpl; auto.
        - rrdes EXT0 EXT0. apply EXT2.
        - rrdes EXT0 EXT0. s in EXT1.
          rrdes EXT'' EXT''. s in EXT''0.
          unnw. eauto.
        - auto.
      }
    + unfold s'; rr; esplits.
      { rr; esplits; eauto.
        - s; rr; esplits; clean; eauto; [|by rr; eauto].
          rewrite pmcells_sng in CPa.
          eapply tgt_confpred_push; eauto.
          unfold mrf'; tmext_simpl; s; auto.
        - dup EXT''; rrdes EXT''; simpl_proj.
          eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
      }
      { rr; esplits; eauto. }
      { s. noemp; rewrite pmplus_swap; f_equal.
        rewrite -!pmplus_assoc; eauto.
      }
      { s; noemp; eauto. }
    + i; vauto.
Grab Existential Variables.
  exact i.
Qed.


Lemma tr_toplvl_correct: forall FI FI' ancha suba subb s rest n va (vb: val) (ty: flextyp) envf
    (EFI: FI' = map_pair id flext_of FI)
    (SUBa: map fst suba = map fst FI)
    (SUBb: map fst subb = map fst FI)
    (TH: is_thunktyp ty)
    (WF: hastype FI' nil vb ty)
    (FUN: ts_db s (ts_fun (ancha + n)) va : Prop)
    (COD: ts_lhp s = {ancha |=> map snd suba} ##
                     {ancha + n |=> map enc (tr_toplvl vb n (map fst FI))} ## rest)
    (ENV: ts_env s = subb ++ envf)
    (VLD: pmvalid (ts_lhp s))
    (NZ: n <> 0),
  @tlsim _ _ ts_gw mylw FI suba subb s va vb ty.
Proof.
  ii; invc WF; destruct vb; destruct e_of; vauto
  ; destruct ty; inv HT; destruct flext_of; vauto.
  - s in FUN; eapply tr_toplvl_correct_fix; eauto.
  - s in FUN; eapply tr_toplvl_correct_gen; eauto.
Qed.


Theorem tr_mod_correct: forall FI FO m FI' FO'
    (EFI: FI' = map_pair id flext_of FI)
    (EFO: FO' = map_pair id flext_of FO)
    (WF: mhastype FI' m FO'),
  msim ts_gw FI FO (tr_mod (map fst FI) m) m.
Proof.
  exists mylw; i.

  destruct ciga, cila; rrdes LMa.
  destruct (load _ _ _) as [[gh lh] exports] eqn:E; des.
  simpl lang_tgt.cpc in *; simpl lang_tgt.crf in *.
  simpl lang_tgt.cst in *; simpl lang_tgt.chp in *.
  clarify.

  destruct cigb, cilb; rrdes LMb; clarify.

  set (db := fun q v =>
         exists v', q = ts_fun v' /\ pmread (hp_imports ## gh) v v').
  set (sig := (rf, db, subb ++ m ++ envf, refst_emp) : state ts_gw).

  set (codeframe := pmemp : heap).
  set (sil := (pmemp, codeframe ## lh) : state mylw).

  exists (sig, sil); splits.
  - ii; vauto.
  - apply mylw_stable.
  - i; rr; econs; exists cemp; econs; exists cemp; splits; noemp; last first; eauto.
    + unfold sig; s; rr; esplits; s; unfold src_cemp, tgt_cemp; eauto
      ; i; try apply pmread_empD in ALC; vauto.
    + rr; splits; econs; s; rr; splits; auto.
      * rewrite GST; split; [by eauto|i; omega].
      * i; des; rrdes IN; subst; s; eauto.
  - rr; esplits; s; unfold codeframe; noemp; eauto.
  - i; unfold sig in *; rrdes FH; vauto.
  - s; unfold tgt_vload; rewrite E; s.
    hexploit load_valid; eauto; []; intro VLDm.

    s in E; clarify; noemp; rewrite ->app_nil_r in *.

    replace (length (g_cptrs _)) with (length m) in *; last first.
    { i; rhs rewrite -(map_length fst) -tr_mod_aux_labels map_length; auto. }

    rename H0 into Eex, H2 into Egh, H1 into Elh.
    rewrite Eex in Elh.
    set (n := 0).
    set (size := 0).

    match goal with H: context[fun p: ?P => ?a + snd p] |- _ =>
      set (relocate := fun p: P => a + snd p) in *;
      assert (relocate = fun p: P => a + n + snd p)
        by (extensionalities; ii; unfold relocate; auto)
    end.
    rewrite ->H in *; clear H relocate.
    match goal with H: context[fun p: ?P => ?a + snd p] |- _ =>
      set (relocate := fun p: P => a + snd p) in *
    end.

    set (Etr := tr_mod_aux _ _ _) in Eex, Egh, Elh.
    assert (Etr = tr_mod_aux (length (map fst suba) + length m + n + length m)
                             size (map fst suba)).
    { unfold Etr, n, size; rewrite SUBa; f_equal; auto. }
    rewrite ->H in *; clear H Etr.
    set (Etr := tr_mod_aux _ _ _ _) in Eex, Egh, Elh.

    replace (length m) with (length m + n) in Egh; [|by auto].
    rewrite plus_assoc in Egh.

    pattern (length m) at 2 in Elh.
    replace (length m) with (length m + n + size) in Elh; [|unfold n, size; omega].
    rewrite !plus_assoc in Elh.

    replace (length m) with (length m + n) in Eex; [|by auto].
    rewrite plus_assoc in Eex.

    assert (VLD: pmvalid (codeframe ## lh)).
    { unfold codeframe; noemp; split_pmvalid LDJ; auto. }
    

    remember (map_pair id flext_of FI) as FI'.
    remember (map_pair id flext_of FO) as FO'.
    destruct WF.
    clear dependent cst; generalize dependent codeframe.
    generalize dependent size; generalize dependent n.
    revert SUBa SUBb.
    i; ginduction MHT; i.
    { s in Etr; s in Eex; subst exports; destruct FO; vauto. }

    hexploit tr_mod_wf.
    { econs; eauto; s in FLO; des; try apply is_thunkenv_app; s; eauto. }
    intro WFM.

    assert (SIM: forall (fty: flextyp) (TT: is_thunktyp fty) (Efty: ty = fty),
                   tlsim FI0 suba subb (sig, sil)
                       (ancha + length (map snd suba) + S (length m) + n) v fty).
    { i; eapply tr_toplvl_correct with (ancha := ancha).
      - auto.
      - auto.
      - auto.
      - destruct fty; vauto.
      - rewrite -Efty -HeqFI'; auto.
      - s; unfold db; esplits; eauto.
        eapply pmread_plus; eauto.
        rewrite -Egh; unfold relocate; s.
        rewrite pmcells_cons.
        rhs rewrite -![in X in X + size]plus_assoc -(plus_assoc _ _ size) pmplus_swap.
        eauto.
      - s; rewrite -Elh pmcells_app -!pmplus_assoc pmplus_swap; symmetry; s.
        rewrite !plus_assoc.
        replace (_ + size) with
          (ancha + length (map snd suba) + S (length m) + S (length m) + n + size)
        ; last first.
        { clear; set (x := length m); set (y:= length _); omega. }
        rewrite pmcells_app HeqFI'.
        replace (length (map snd suba) + _ + size)
           with (length (map fst suba) + S (length m) + n + S (length m) + size)
        ; last first.
        { clear; set (x := length m); rewrite !map_length; omega. }
        rewrite SUBa.
        (rhs rewrite -2!(pmplus_swap {ancha + _ + _ + _ + _ + _ |=> _})); eauto.
      - s; eauto.
      - eauto.
      - omega.
    }

    destruct FO0; vauto.
    destruct p; s in HeqFO'; clarify.
    unfold id in H2; subst l0.  

    apply list_assoc_cons_or in IN; des.
    + 
      s in Eex.
      subst f f0; esplits.
      * rewrite -Eex list_assoc_cons_eq; eauto.
      * rr; s; des_if; vauto.
      * apply SIM; eauto.
        s in FLO; des; destruct ty0; vauto; destruct flext_of; vauto; destruct ty; vauto.
    + 
      
      hexploit (IHMHT (FI0 ++ [(l, f0)]) FO0).
      * eauto.
      * eauto.
      * instantiate (1 := hp_imports ## {ancha + length (map snd suba) + S (length m) + n |-> relocate (l, size)}).
        instantiate (1 := suba ++ [(l, ancha + length (map snd suba) + S (length m) + n)]).
        rewrite -Egh in GDJ; s in GDJ; rewrite pmcells_cons in GDJ.
        split; i.
        { decomp_pmvalid GDJ.
          rewrite map_app in H; apply in_app_or in H; des.
          - apply GDOM in H; des; econs; eapply pmread_plus; eauto.
          - s in H; des; vauto; rewrite -H.
            econs; eapply pmread_plus; eauto; try pmplus_eq'.
        }
        { des; apply pmread_plusD in H; des.
          - rewrite map_app; apply in_or_app; left.
            apply GDOM; eauto.
          - rewrite map_app; apply in_or_app; right.
            apply pmread_cellD in H; des; s; eauto.
        }
      * rewrite -Egh in GDJ; rewrite -Elh in LDJ.
        rewrite pmcells_cons in GDJ; s in GDJ; decomp_pmvalid GDJ.
        s in LDJ; rewrite (pmcells_app (map enc _)) in LDJ.
        revert LDJ.
        match goal with |- context[hp_imports ## ?a ## ?b ## ?c] =>
          instantiate (1 := a ## c)
        end.
        i; decomp_pmvalid LDJ.
        rewrite -!pmplus_assoc.
        apply pmvalid_plus_plus.
        eauto.
        eapply pmvalid_plus_plus; eauto.
        rewrite pmplus_swap.
        eapply pmvalid_plus_plus; eauto.
        { rewrite -Egh -Elh in VLDm.
          s in VLDm; rewrite pmcells_cons -!pmplus_assoc in VLDm.
          decomp_pmvalid VLDm.
          rewrite pmplus_comm; eauto.
        }
        rewrite -Egh -Elh in VLDm.
        s in VLDm; rewrite pmcells_cons !pmcells_app -!pmplus_assoc in VLDm.
        decomp_pmvalid VLDm; eauto.
      * rewrite -Egh in GDJ; s in GDJ; rewrite pmcells_cons in GDJ.
        rewrite <-pmplus_assoc; eauto.
      * rewrite map_pair_app HeqFI' H1; auto.
      * eauto.
      * s in FLO; des; apply is_thunkenv_app; s; auto.
      * s in FLO; des; auto.
      * rewrite pmplus_swap.
        rewrite -Egh -Elh in VLDm.
        s in VLDm; rewrite pmcells_cons -!pmplus_assoc in VLDm.
        rewrite (pmcells_app _ (g_data _)) in VLDm.
        decomp_pmvalid VLDm.
        apply pmvalid_plus_plus; eauto.
        rewrite pmplus_comm; eauto.
      * s; rewrite !map_app SUBa; auto.
      * instantiate (1 := subb ++ [(l, v)]).
        rewrite !map_app SUBb; auto.
      * instantiate (6 := ancha).
        s; rewrite !map_length !app_length. s.
        instantiate (2:=S n). f_equal.
        i. unfold tgt_val. omega. 
        f_equal.
        { unfold relocate; extensionalities.
          s. rewrite map_length. unfold tgt_val; omega.
        }
        f_equal; f_equal; unfold tgt_val. omega.
        rewrite map_app; s; auto.
      * f_equal.
        { rewrite ->map_app, <-app_assoc.
          rewrite -Eex. s. eauto.
        }
        f_equal.
        { rewrite !map_length !app_length. s. unfold tgt_val. omega. }
        f_equal; f_equal.
        rewrite map_app.
        rewrite app_length; s; omega.
        rewrite map_length; auto.
        rewrite map_app; s; auto.
      * s; f_equal; f_equal. rewrite !map_length app_length. s. unfold tgt_val. omega.
        f_equal.
        f_equal.
        f_equal.
        rewrite ?map_length !app_length. s. unfold tgt_val. omega.
        rewrite map_length; auto.
        rewrite map_app; s; auto.
      * rewrite -Elh in VLD; s in VLD.
        rewrite (pmcells_app (map enc _)) (pmplus_swap {ancha |=> _}) in VLD.
        revert VLD.
        match goal with |- context[codeframe ## ?a ## ?b ## ?c] =>
          instantiate (1 := codeframe ## a)
        end.
        i; rewrite -!pmplus_assoc; eauto.
      * eauto.
      * i; des; esplits.
        { rewrite -Eex; s; des_if; [by exfalso; ii; subst f; vauto|].
          eauto.
        }
        { unfold src_vload.
          rewrite list_assoc_cons_neq; vauto.
          ii; subst f; vauto.
        }
        ii; eapply H3; eauto.
        { unfold sig, db, sil in LE.
          rewrite -Elh -Egh in LE. simpl length in LE.
          rewrite -!pmplus_assoc -(pmplus_swap {ancha |=> _}).
          simpl g_cptrs in LE.
          rewrite pmcells_cons (pmcells_app (map enc _)) in LE.
          fold tr_mod_aux in LE.
          simpl length in LE.
          rewrite -app_assoc -cons_app.
          rewrite -app_comm_cons in LE; eauto.
        }
        i; apply list_assoc_app_or in INsuba.
        apply list_assoc_app_or in INsubb; des.
        { apply list_assoc_app_or in INienv; des; [by eauto|].
          exfalso. apply list_assoc_someD in INsuba.
          eapply (in_map fst) in INsuba.
          rewrite SUBa in INsuba.
          eapply (list_assoc_map eqdec) in INsuba; des.
          rewrite INienv in INsuba; vauto.
        }
        { apply list_assoc_someD in INsubb.
          apply list_assoc_noneD in INsuba; vauto.
          rewrite SUBa -SUBb; eapply (in_map fst) in INsubb; auto.
        }
        { apply list_assoc_someD in INsuba.
          apply list_assoc_noneD in INsubb; vauto.
          rewrite SUBb -SUBa; eapply (in_map fst) in INsuba; auto.
        }
        { apply list_assoc_someD in INsuba0; apply list_assoc_someD in INsubb0.
          s in INsuba0; s in INsubb0; des; vauto; subst f1 va' vb'.
          assert (ty = ty1).
          { apply list_assoc_app_or in INienv; des
            ; last (by s in INienv0; des_ifs; subst f0; eauto).
            exfalso; apply list_assoc_someD in INienv.
            eapply (in_map fst) in INienv.
            rewrite -SUBa in INienv.
            eapply (list_assoc_map eqdec) in INienv; des.
            rewrite INienv in INsuba; vauto.
          }
          subst ty; s in FLO; des.
          eapply vclos_thunk_mon; first (eapply SIM); eauto using (gk_lsim GK).
          rewrite -H4; auto.
        }
Grab Existential Variables.
  exact tt.
Qed.
