(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw_common gw ssr util ts.
Require Import src2tgt src2tgt_correct_aux.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Lemma tr_exp_correct_refeq: forall F G ty e e'
    (IH: ts_psim F G (Tref ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (IH': ts_psim F G (Tref ty) (fun n => map enc (tr_exp e' n (map fst F) (map fst G))) e'),
  ts_psim F G Tnat (fun n => map enc (tr_exp (Erefeq e e') n (map fst F) (map fst G))) (Erefeq e e').
Proof.
  i; rdes IH; rdes IH'; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKrefeqL _ ::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.

  destruct (LEM (pmvalid (snd (snd s')))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH' with (kb:= SKrefeqR _ ::kb); clear IH'; eauto using ext_trans.
  { i; apply RSV in IN; des; esplits; eauto.
    - rrdes EXT0 EXT0 EXT0 EXT0; rewrite EXT3; eauto.
    - eapply vclos_mon; eauto using (gk_mon GK).
  }
  { i; apply RDV in IN; des; esplits; eauto.
    eapply vclos_mon; eauto using (gk_mon GK).
  }
  { rewrite ENV0.
    rrdes EXT0 EXT0 EXT0 EXT0.
    rrdes EXT1; eapply reprlist_mon; eauto.
  }
  { omega. }
  { rrdes EXT0 EXT0 EXT0 EXT0; rrdes EXT1.
    hexploit (pmplus_from_read EXT1); eauto.
    instantiate (1 := length (map enc (tr_exp e n (map fst F) (map fst G))) + n).
    i; des; rewrite H Eladdr !map_app !pmcells_app plus_comm.
    econs; symmetry.
    replace (_ - _) with (laddr - n); last first.
    { clear; set (x := length _); omega. }
    rewrite <-!pmplus_assoc.
    rhs rewrite pmplus_swap. rewrite !map_length.
    rewrite ->pmplus_swap; pmplus_eq'.
  }
  i; ts_prepare.

  
  inv VR; invc REL; [|by invc VR1]; rrdes VR1.
  apply EXT1 in REG.
  inv VR0; invc REL; [|by invc VR1]; rrdes VR1.

  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { rewrite IRb0 IRb; econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; s; noemp; rr; esplits; eauto. }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.
  - assert (EXT2: ext _ s'0 (r0, P0, s, s3, s4))
      by (eapply ext_trans; eauto).

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc EXT2 Eladdr laddr.
        unfold pop; auto. 
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc EXT2 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        s; rewrite LST0 LST SP0 SP.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
        (rhs rewrite -!pmplus_assoc -!(pmplus_swap {r0 r_sp |-> _})); auto.
      - econs; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        s; rewrite LST0 LST SP0 SP.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
        (rhs rewrite -!pmplus_assoc pmplus_swap); auto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT2 Eladdr laddr.
      - unfold stack_top; rewrite LST0 LST SP0 -!pmplus_assoc.
        rewrite -!(pmplus_swap {_|->va}).
        replace (_ r_sp - 1) with (fst (fst (fst (fst s'0))) r_sp - 1 - 1)
        ; [|rewrite SP0; omega].
        prep_store; eauto.
    }
    { s; noemp; tmext_simpl.
      rewrite LST0 LST SP0 -!pmplus_assoc pmplus_swap in SV0.
      depmcplus0; eauto using pmvalid_plus_cell.
      rewrite SP0.
      repeat match goal with
        | |- context[?a + 1 - 1] => replace (a + 1 - 1) with a by omega
        | H: context[?a + 1 - 1] |- _ => replace (a + 1 - 1) with a in H by omega
      end; eauto.
      rewrite (pmplus_swap freest) in SV0.
      rewrite -!(pmplus_swap {_|->va0}).
      eapply pmvalid_plus_cell.
      rewrite pmplus_swap in SV0; rewrite pmplus_swap.
      eapply pmvalid_plus_cell; eauto.
    }

    eapply dance_O.
    instantiate (1 := fun _ _ => depmach0); s; eauto.

  - destruct s'0 as [[[[]]]]; clean.
    i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.

      set (s'' := (mrf, P1, env, s6,
                   ({r1 r_sp - 1 - 1 |-> if eq_nat_dec va va0 then 1 else 0} ## fst s4, snd s7))).

      assert (EXT'': ext (wlift mylw) s'' (r1, P1, env, s6, s7)).
      { clear -HV ; unfold s''; s; splits; tmext_simpl; auto. }
      exists s''; esplits; [exact EXT''|].

      hexploit (KSIM g'0).
      * eauto.
      * eauto.
      * repeat (eapply ext_trans; eauto).
      * s; unfold mrf; tmext_simpl; congruence.
      * s; unfold mrf; tmext_simpl.
        rewrite SP0 SP; omega.
      * s; unfold mrf; tmext_simpl; congruence.
      * s; unfold mrf; tmext_simpl; auto.
      * s; econs; econs; rr; eauto.
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            eapply tgt_confpred_pop with (popped := [va0]); eauto.
            + unfold mrf; tmext_simpl; auto.
            + rewrite -pmcells_sng; auto.
              unfold mrf; tmext_simpl.
              rewrite pmplus_swap SP !(pmplus_swap freest) in SV.
              rewrite SP0 SP.
              decomp_pmvalid SV; eauto using pmvalid_plus_cell.
          - dup EXT''; rrdes EXT''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp.
        i; eapply pesim_rewrite; eauto.
        f_equal.
        { f_equal; rewrite !map_length !app_length; s; omega. }
        { unfold mrf; tmext_simpl.
          rewrite SP0 SP -!pmplus_assoc; noemp.
          repeat match goal with |- context[?a + 1 - 1] =>
            replace (a + 1 - 1) with a by omega
          end.
          rewrite -!(pmplus_swap {_|->va0}) (pmplus_swap freest); auto.
        }
        { rewrite -!pmplus_assoc; auto. }
    + rr; esplits; s; eauto.
      rewrite -plus_n_O.
      hexploit (refst_bij REG REG0); []; i; des.
      des_ifs; specialize (X eq_refl) || specialize (X0 eq_refl); des; subst; vauto.
Grab Existential Variables.
  exact (wfo_nat 42).
Qed.



Lemma sto_hind': forall (pc : word) (rf : tmap word) (st : stack) 
    (h : pmap word) (r : reg) (soff : signed_offset) h'
    (v w : word)
    (E: h' = {add_offset (rf r) soff |-> v} ## h),
  store pc rf st h' (o_hind r soff) w
        {| mpc := 1 + pc; mrf := rf; mst := st;
           mhp := {add_offset (rf r) soff |-> w} ## h |}.
Proof.
  i; rewrite E; eauto.
Qed.                            



(** Convenience lemmas for reasoning about reference allocation. *)

Lemma ts_refconfrel_newa: forall {g mrf P0 s s3 refha refhb la va}
    (RCR: ts_refconfrel g (mrf, P0, s, s3)
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := refha |}
            (refhb, Nothing, Nothing))
    (VLD: pmvalid ({la |-> va} ## refha)),
  forall ty lb', ~ s3 ty la lb'.
Proof.
  i; rrdes RCR; clarify.
  apply RELV in FH; des; rrdes RDa; subst.
  rewrite pmplus_assoc in VLD; apply pmvalid_plus_l in VLD.
  apply pmvalid_cell_cell in VLD; vauto.
Qed.

Lemma ts_refconfrel_newb: forall {g mrf P0 s s3 refha refhb lb vb}
    (RCR: ts_refconfrel g (mrf, P0, s, s3)
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := refha |}
            (refhb, Nothing, Nothing))
    (VLD: pmvalid ({lb |-> vb} ## refhb)),
  forall ty la', ~ s3 ty la' lb.
Proof.
  i; rrdes RCR; clarify.
  apply RELV in FH; des; rrdes RDb; subst.
  rewrite pmplus_assoc in VLD; apply pmvalid_plus_l in VLD.
  apply pmvalid_cell_cell in VLD; vauto.
Qed.

Lemma ts_refconfrel_alloc: forall g mrf P0 s s3 s4 ha hb va vb la lb mrf' P0' s' s3' s4' ty
    (RCR: ts_refconfrel (fun sg => g (sg, s4)) (mrf, P0, s, s3)
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := ha |}
            (hb, Nothing, Nothing))
    (VCL: vclos (wlift mylw) (mrf, P0, s, s3, s4) (g (mrf, P0, s, s3, s4)) ty va vb)
    (EXT: ext (wlift mylw) (mrf', P0', s', s3', s4') (mrf, P0, s, s3, s4))
    (GK: gknow (wlift mylw) g)
    (ALC: refst_alloc s3 ty la lb = s3')
    (VLDa: pmvalid ({la |-> va} ## ha))
    (VLDb: pmvalid ({lb |-> vb} ## hb)),
  ts_refconfrel (fun sg => g (sg, s4')) (mrf', P0', s', s3')
    {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := {la |-> va} ## ha |}
    ({lb |-> vb} ## hb, Nothing, Nothing).
Proof.
  unfold refst_alloc; i.
  rrdes RCR; clarify; clean.
  rr; esplits; eauto; last first; clean.
  - rewrite -ALC; i; des.
    + hexploit RELV; eauto; i; des; esplits
      ; try by (eapply pmread_plus; eauto; pmplus_eq').
      dup EXT; rrdes EXT; eapply vclos_mon; eauto using (gk_mon GK).
    + subst; esplits
      ; try by (eapply pmread_plus; eauto; pmplus_eq').
      eapply vclos_unlift in VCL.
      dup EXT; rrdes EXT; eapply vclos_mon; eauto using (gk_mon GK).
  - rewrite -ALC; i; apply pmread_plusD in ALC0; des.
    + apply pmread_cellD in ALC0; des; s; eauto 10.
    + hexploit DOMb; eauto; i; des; s; eauto.
  - rewrite -ALC; i; apply pmread_plusD in ALC0; des.
    + apply pmread_cellD in ALC0; des; s; eauto 10.
    + hexploit DOMa; eauto; i; des; s; eauto.
Qed.


(** Convenvience lemmas for reasoning about reference writing. *)

Lemma ts_refconfrel_write: forall (g : state (wlift mylw) -> vrelf) s ty la lb va vb ha hb
    (RCR: ts_refconfrel (fun sg => (g (sg, snd s))) (fst s)
            {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := {la |-> va} ## ha |}
            ({lb |-> vb} ## hb, Nothing, Nothing))
    (RL: snd (fst s) ty la lb)
    va' vb' (VCL: vclos ts_gw (fst s) (g s) ty va' vb'),
  ts_refconfrel (fun sg => (g (sg, snd s))) (fst s)
    {| cpc := Nothing; crf := Nothing; cst := pmemp; chp := {la |-> va'} ## ha |}
    ({lb |-> vb'} ## hb, Nothing, Nothing).
Proof.
  i; rrdes RCR; clarify; rr; esplits; eauto using pmvalid_plus_cell.
  - i; dup ALC; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; subst; eauto.
    + hexploit DOMa; [|by eauto].
      rrdes ALC0; eapply pmvalid_plus_cell in VLD.
      eapply pmread_plus; eauto; pmplus_eq'.
  - i; dup ALC; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; subst; eauto.
    + hexploit DOMb; [|by eauto].
      rrdes ALC0; eapply pmvalid_plus_cell in VLD.
      eapply pmread_plus; eauto; pmplus_eq'.
  - i; dup REG; apply RELV in REG; des.
    dup RDa; rrdes RDa; dup RDb; rrdes RDb.
    apply pmread_plusD in RDa0; apply pmread_plusD in RDb0; des.
    + apply pmread_cellD in RDa0; apply pmread_cellD in RDb0; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X0 eq_refl); des; subst.
      rewrite -prod_eta; esplits; last (exact VCL)
      ; eauto using pmread_plus, pmread_cell, pmvalid_plus_cell.
    + exfalso.
      apply pmread_cellD in RDb0; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X0 eq_refl); des; subst.
      rrdes RDa0; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLD.
      apply pmvalid_plus_l in VLD.
      apply pmvalid_cell_cell in VLD; vauto.
    + exfalso.
      apply pmread_cellD in RDa0; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X eq_refl). subst.
      rrdes RDb0; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLD0.
      apply pmvalid_plus_l in VLD0.
      apply pmvalid_cell_cell in VLD0; vauto.
    + eauto 10 using pmread_plus, pmplus_comm, pmvalid_plus_cell.
Qed.                            



Lemma tr_exp_correct_write: forall F G ty e e'
    (IH: ts_psim F G (Tref ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (IH': ts_psim F G ty (fun n => map enc (tr_exp e' n (map fst F) (map fst G))) e'),
  ts_psim F G Tunit (fun n => map enc (tr_exp (Ewrite e e') n (map fst F) (map fst G))) (Ewrite e e').
Proof.
  i; rdes IH; rdes IH'; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKwriteL _ ::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.

  destruct (LEM (pmvalid (snd (snd s')))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH' with (kb:= SKwriteR _ ::kb); clear IH'; eauto using ext_trans.
  { i; apply RSV in IN; des; esplits; eauto.
    - rrdes EXT0 EXT0 EXT0 EXT0; rewrite EXT3; eauto.
    - eapply vclos_mon; eauto using (gk_mon GK).
  }
  { i; apply RDV in IN; des; esplits; eauto.
    eapply vclos_mon; eauto using (gk_mon GK).
  }
  { rewrite ENV0.
    rrdes EXT0 EXT0 EXT0 EXT0.
    rrdes EXT1; eapply reprlist_mon; eauto.
  }
  { omega. }
  { rrdes EXT0 EXT0 EXT0 EXT0; rrdes EXT1.
    hexploit (pmplus_from_read EXT1); eauto.
    instantiate (1 := length (map enc (tr_exp e n (map fst F) (map fst G))) + n).
    i; des; rewrite H Eladdr !map_app !pmcells_app plus_comm.
    econs; symmetry.
    replace (_ - _) with (laddr - n); last first.
    { clear; set (x := length _); omega. }
    rewrite <-!pmplus_assoc.
    rhs rewrite pmplus_swap. rewrite !map_length.
    rewrite ->pmplus_swap; pmplus_eq'.
  }
  i; ts_prepare.

  
  inv VR; invc REL; [|by invc VR1]; rrdes VR1.
  apply EXT1 in REG.
  dup WHY0; rrdes WHY0; hexploit RELV; eauto; i; des.
  rrdes RDa RDb; clarify; rewrite <-!pmplus_assoc in *.

  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  s in RL; rewrite -pmplus_assoc in RL.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { rewrite IRb; econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; instantiate (3 := _ ## _).
    s; noemp; rewrite pmplus_assoc; rr; esplits; eauto.
    - rewrite -pmplus_assoc; eauto using pmvalid_plus_cell. 
    - rewrite -pmplus_assoc; eauto using pmfinite_plus_cell. 
  }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.
  - assert (EXT2: ext _ s'0 (r0, P0, s, s3, s4))
      by (eapply ext_trans; eauto).

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc EXT2 Eladdr laddr.
        unfold pop; auto. 
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT2 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        symmetry; rewrite pmplus_swap LST0 LST SP0 SP.
        rewrite -!pmplus_assoc pmplus_swap.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT2 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        symmetry; rewrite pmplus_swap LST0 LST SP0 SP.
        rewrite -!pmplus_assoc.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT2 Eladdr laddr.
      - rewrite ->(pmplus_swap hp).
        
        eapply sto_hind'.
        s; tmext_simpl; auto.
    }
    { s; tmext_simpl.
      rewrite pmplus_swap in HV0, FIN0.
      depmcplus0; eauto using pmvalid_plus_cell, pmfinite_plus_cell.
    }

    eapply dance_O.
    instantiate (1 := fun _ _ _ => depmach0); s; eauto.

  - destruct s'0 as [[[[]]]]; clean.
    i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.

      set (s'' := (mrf, P1, env, s6, ({r1 r_sp - 1 - 1 |-> va} ## fst s4, snd s7))).

      assert (EXT'': ext (wlift mylw) s'' (r1, P1, env, s6, s7)).
      { clear -HV ; unfold s''; s; splits; tmext_simpl; auto. }
      exists s''; esplits; [exact EXT''|].

      hexploit (KSIM g'0).
      * eauto.
      * eauto.
      * repeat (eapply ext_trans; eauto).
      * s; unfold mrf; tmext_simpl; congruence.
      * s; unfold mrf; tmext_simpl.
        rewrite SP0 SP; omega.
      * s; unfold mrf; tmext_simpl; congruence.
      * s; unfold mrf; tmext_simpl; auto.
      * s; econs; econs; rr; eauto.
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            eapply tgt_confpred_pop with (popped := [va0]); eauto.
            + unfold mrf; tmext_simpl; auto.
            + rewrite -pmcells_sng; auto.
              rewrite LST0 pmplus_swap in SV.
              unfold mrf; tmext_simpl.
              decomp_pmvalid SV; eauto using pmvalid_plus_cell.
          - inst_pairs; eapply ts_refconfrel_write; eauto.
            + eapply ts_refconfrel_mon
              ; [eapply ts_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto; rrdes EXT''; rr; eauto.
            + eapply vclos_unlift, vclos_mon; eauto using (gk_mon GK'0).
        }
        simpl cplus; noemp.
        i; eapply pesim_rewrite; eauto.
        f_equal.
        { f_equal; rewrite !map_length !app_length; s; omega. }
        { unfold mrf; tmext_simpl; rewrite LST0.
          rewrite -!pmplus_assoc; noemp.
          rewrite LST SP0 SP pmplus_swap.
          do 4 f_equal; omega.
        }
        { rewrite -!pmplus_assoc pmplus_swap; auto. }
    + rr; esplits; s; eauto.
Grab Existential Variables.
  exact (wfo_nat 42).
Qed.



Lemma tr_exp_correct_read: forall F G ty e
    (IH: ts_psim F G (Tref ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G ty (fun n => map enc (tr_exp (Eread e) n (map fst F) (map fst G))) (Eread e).
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ (wfo_nat 7)); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKread::kb); eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  i; ts_prepare.

  
  inv VR; invc REL; [|by invc VR0]; rrdes VR0.
  dup WHY0; rrdes WHY0; hexploit RELV; eauto; i; des.

  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { rewrite IRb; econs 1; inst_pairs; eapply Rok.
    eapply red_ctx; econs; eapply Rread.
    eapply pmread_plus; eauto.
  }
  { inst_pairs; s; noemp; rr; esplits; eauto. }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.
  -
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        (rhs (rewrite LST -pmplus_assoc pmplus_swap)); eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        (rhs rewrite pmplus_swap); eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - unfold stack_top; rewrite LST -!pmplus_assoc pmplus_swap.
        prep_store; eauto.
    }
    { s; noemp; tmext_simpl.
      rewrite LST -!pmplus_assoc pmplus_swap in SV0.
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    eapply dance_O.
    instantiate (1 := fun _ _ => depmach0); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.

      set (s' := (mrf, P1, env, s6, ({r1 r_sp - 1 |-> va0} ## fst s4, snd s7))).

      assert (EXT': ext (wlift mylw) s' (r1, P1, env, s6, s7)).
      { clear -HV ; unfold s'; s; splits; tmext_simpl; auto. }
      exists s'; esplits; [exact EXT'|].

      hexploit (KSIM g').
      * eauto.
      * eauto.
      * eapply ext_trans; eauto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * dup EXT'; rrdes EXT'.
        eapply vclos_lift, vclos_mon; eauto.
        i; eapply gk_mon; eauto.
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            insts $ {| cpc := Nothing; crf := Just mrf; cst := freest; chp := hp |}.
            unfold mrf; rrdes CPa; rr; s; tmext_simpl; esplits; eauto.
          - dup EXT'; rrdes EXT'; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp.
        rewrite -!pmplus_assoc (pmplus_swap freest) (pmplus_swap hp).
        esplits; last first; eauto.
    + rr; esplits; s; eauto.
Qed.



Lemma tr_exp_correct_new: forall F G ty e
    (IH: ts_psim F G ty (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G (Tref ty) (fun n => map enc (tr_exp (Enew e) n (map fst F) (map fst G))) (Enew e).
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ (wfo_nat 7)); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKnew::kb); eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  i; ts_prepare.

  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify.
  dup PV0; eapply (pmfinite_alloc vb) in PV0; eauto; des.
  esplits; [|right; esplits].
  { econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; instantiate (3 := _ ## _).
    s; noemp; rewrite pmplus_assoc; rr; esplits; eauto.
    - rewrite -pmplus_assoc; eauto. 
    - rewrite -pmplus_assoc; eauto. 
  }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.
  -
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
      readpc EXT0 Eladdr laddr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        (rhs (rewrite LST -pmplus_assoc pmplus_swap)); eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - unfold stack_top; rewrite LST -!pmplus_assoc pmplus_swap.
        prep_store; eauto.
    }
    { s; noemp; tmext_simpl.
      rewrite LST -!pmplus_assoc pmplus_swap in SV0.
      rewrite -pmcells_sng in HV0, FIN0.
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    eapply dance_O.
    instantiate (1 := fun _ cf _ _ _ => depmach1 cf); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.

      assert (VLDa: pmvalid ({newloc |-> va} ## refha)).
      { decomp_pmvalid HV; eauto. }

      assert (VLDb: pmvalid ({x |-> vb} ## h1)).
      { decomp_pmvalid PV0; eauto. } 

      set (s' := (mrf, P1, env,
                  refst_allocP s6 (tsubst_name tsub ty) newloc x
                               (ts_refconfrel_newa WHY0 VLDa)
                               (ts_refconfrel_newb WHY0 VLDb),
                  ({r1 r_sp - 1 |-> newloc} ## fst s4, snd s7))).

      assert (EXT': ext (wlift mylw) s' (r1, P1, env, s6, s7)).
      { clear -HV ; unfold s'.
        unfold refst_alloc; s; splits; tmext_simpl; auto.
        unfold refst_alloc; eauto.
      }
      exists s'; esplits; [exact EXT'|].

      hexploit (KSIM g').
      * eauto.
      * eauto.
      * eapply ext_trans; eauto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * simpl tsubst_name.
        econs; eapply vclgen_ref; rr; esplits; eauto.
        unfold s'; s; rr; eauto.
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            insts $ {| cpc := Nothing; crf := Just mrf; cst := freest; chp := hp |}.
            unfold mrf; rrdes CPa; rr; s; tmext_simpl; esplits; eauto.
          - inst_pairs; eapply ts_refconfrel_alloc; s; eauto.
        }
        simpl cplus; noemp.
        rewrite -!pmplus_assoc (pmplus_swap freest) (pmplus_swap hp).
        esplits; last first; eauto.
    + rr; esplits; s; eauto.
Qed.



Lemma tr_exp_correct_input: forall F G,
  ts_psim F G Tnat (fun n => map enc (tr_exp Einput n (map fst F) (map fst G))) Einput.
Proof.
  i; econs; red; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ (wfo_nat 1)); rewrite <-wfo_add_assoc.
  i; ts_prepare.

  

  
  
  
  
  
  
  
  
  
  

  
  

  
  
  
  
  
  

  
  
  
  

  pfold; econs; i; esplits; eauto.
  destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
  destruct cfb; rrdes XTRb; clarify.
  invc CNF; des; subst; noemp.
  rrdes MB; s in Ec; clarify.
  destruct ma; clear SV HV FIN; simpl cplus in MA.
  destruct_tgt_real MA.
  rewrite <-!pmplus_assoc in *; clarify.

  assert (INP: pmread (hp ## refha ## snd s4 ## chp) mpc0 (enc (i_inp r_clo))).
  { readpc' mpc0 Eladdr. }

  eapply esim_step; first (i; rrdes FH; ss; omega); i.
  inv ST; last by (exfalso; eapply ST0, Sinp; eauto).
  inv ST0; hexploit (pmread_uniq INS INP); intro ENC
  ; apply enc_inj in ENC; vauto.

  set (s' := ([n0//r_clo]>>mrf0, P0, env, s3, s4)).
  assert (EXT': ext (wlift mylw) s' (mrf0, P0, env, s3, s4)).
  { rr; s; esplits; eauto. }

  esplits; first (by eauto); last (exact EXT'); [left; econs; eauto|]
  ; last first; last 1 first.
  { left; tmext_simpl.
    rr; inst_pairs; apply Rok, red_ctx; eauto.
  }
  { econs; eauto.
    + instantiate (1:=mkconf _ _ _ _); s; noemp; eauto.
      rr; esplits; s; eauto; pmplus_eq.
    + inst_pairs; s; noemp; eauto.
      rr; esplits; s; eauto; pmplus_eq.
  }
  rewrite clear_bote; eapply gres_incl_id.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc in *; clear NST.

  eapply pesim_dance.
  - rrdes RLa0; instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto.
  - clear SV HV FIN FIN0 ST ST0 NZ NZ0 INS INP MA0 RLa0.
    s; setmrf mrf.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc' mpc0 Eladdr.    
        unfold push; auto. 
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc0 Eladdr.
      - unfold stack_top; rewrite -pmcells_sng.
        prep_store; eauto.
    }
    { s; tmext_simpl.
      match goal with |- context[?a + 1 - 1] =>
        replace (a + 1 - 1) with a by omega
      end.
      depmcplus0; rewrite pmcells_sng.
      eauto using pmvalid_plus_cells.
    }

    eapply dance_O.
    instantiate (1 := fun _ => depmach0); s; eauto.

  - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      unfold s' in *; clear s'.
      destruct ea'; clear FIN ST SV HV; destruct_tgt_real RLa.
      clear RLa0.
      setcrf mrf'.
      set (s4' := ({mrf0 r_sp |-> n0} ## fst s4, snd s4)).
      set (s' := (mrf', P0, env0, s3, s4')).
      assert (EXT'': ext (wlift mylw) s' ([n0//r_clo]>>mrf0, P0, env0, s3, s4)).
      { unfold s'; repeat (split; auto); unfold mrf'; s; tmext_simpl; auto.
        rrdes IN; eauto.
      }
      esplits; eauto.

      hexploit (KSIM g); clear KSIM.
      * eauto.
      * eauto.
      * exact EXT''.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl.
        match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end.
        auto.
      * s; econs; econs; rr; eauto.
      * intro X; simpl tr_exp in X.
        exploit X; clear X.
        { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto]; s.
            eapply (tgt_confpred_push mrf'); eauto.
            unfold mrf'; s; tmext_simpl; auto.
          - dup EXT''; rrdes EXT''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        rewrite pmplus_swap; f_equal.
        rewrite -pmplus_assoc; auto.
    + rewrite <-H0, <-H2, <-H1, <-plus_n_O in *; s; noemp; rr; esplits; eauto.
Grab Existential Variables.
  exact 42. exact (wfo_nat 42).
Qed.


Lemma tr_exp_correct_output: forall F G e
    (IH: ts_psim F G Tnat (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G Tunit (fun n => map enc (tr_exp (Eoutput e) n (map fst F) (map fst G))) (Eoutput e).
Proof.
  i; rdes IH; econs; red; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ (wfo_nat 1)); rewrite <-wfo_add_assoc.
  
  eapply IH with (kb:= SKoutput::kb); eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  i; ts_prepare.

  

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.
  - dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
    }
    { depmcplus0. }

    eapply dance_O.
    instantiate (1 := depmach0); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb0; clarify.
    s in Ec; noemp; clarify.
    esplits; first 1 last.
    { right; splits; eauto using wfo_le_n_add. }
    { inst_pairs; s; noemp; rr; esplits; eauto. }

    i; frames_nodep RLa cfas.
    destruct ea'; clear FIN; destruct_tgt_real RLa.
    setcrf mrf.
    esplits; eauto.

    pfold; econs; i; esplits; eauto.
    destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
    destruct cfb; rrdes XTRb; clarify.
    invc CNF; des; subst; noemp.
    rrdes MB; s in Ec; clarify.
    destruct ma; clear SV HV FIN; simpl cplus in MA.
    destruct_tgt_real MA.
    rewrite <-!pmplus_assoc in*. clarify.

    invc VR REL; [|by inv VR]; rrdes CA CB; invc SVQ.

    assert (OUT: pmread (hp ## refha ## snd s7 ## chp)
                        (S (length (tr_exp e n (map fst F) (map fst G)) + laddr))
                        (enc (i_out r_clo))).
    { readpc EXT0 Eladdr laddr. }

    eapply esim_step; first (i; rrdes FH; ss; omega); i.
    rewrite map_length in ST; inv ST
    ; last by (exfalso; eapply ST0, Sout; eauto).
    inv ST0; hexploit (pmread_uniq INS OUT); intro ENC
    ; apply enc_inj in ENC; vauto.

    set (s' := ([n0//r_clo]>>r1, P1, env0, s6, s7)).
    assert (EXT': ext (wlift mylw) s' (r1, P1, env0, s6, s7)).
    { rr; s; esplits; eauto. }

    esplits; first (by eauto); last (exact EXT'); [left; econs; eauto|]
    ; last first; last 1 first.
    { left; tmext_simpl.
      rr; inst_pairs; apply Rok, red_ctx; eauto.
    }
    { econs; eauto.
      + instantiate (1:=mkconf _ _ _ _); s; noemp; eauto.
        rr; esplits; s; eauto; pmplus_eq.
      + inst_pairs; s; noemp; eauto.
        rr; esplits; s; eauto; pmplus_eq.
    }
    rewrite clear_bote; eapply gres_incl_id.

    hexploit (KSIM g'); clear KSIM.
    { eauto. }
    { eauto. }
    { eapply ext_trans; eauto. }
    { s; unfold mrf; tmext_simpl.
      rewrite ENV0; auto. }
    { s; unfold mrf; tmext_simpl.
      rewrite SP; omega. }
    { s; unfold mrf; tmext_simpl; eauto. }
    { s; unfold mrf; tmext_simpl; eauto. }
    { econs; econs; rr; eauto. }
    intro X; simpl tr_exp in X.
    rewrite map_app app_length in X; simpl length in X.
    setcrf mrf.
    exploit X; clear X.
    { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
      rr; esplits; eauto.
      - rr; splits; [|by rr; esplits; eauto].
        instantiate (1 := {| cpc := Nothing; crf := Just mrf
                             ; cst := freest; chp := hp |}).
        unfold mrf; rrdes CPa; rr; esplits; eauto.
        s; tmext_simpl; eauto.
      - dup EXT'; rrdes EXT'.
        eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
        s; i; eauto using (gk_mon GK').
    }

    simpl cplus; noemp; intro ES.
    eapply pesim_rewrite; eauto.
    rewrite map_length -(plus_comm 2) -!pmplus_assoc; eauto.
Qed.
