(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw_common gw ssr util ts.
Require Import zwickel zwickel_correct_aux.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Lemma tr_exp_correct_fst: forall F G ty ty' e
    (IH: ts_psim F G (Tprod ty ty') (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G ty (fun n => map enc (tr_exp (Efst e) n (map fst F) (map fst G))) (Efst e).
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  eapply IH with (kb:= SKfst::kb); eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  i; ts_prepare.

  invc VR REL; [|by inv VR]; rrdes CB; invc SVQ.
  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; s; noemp; rr; esplits; eauto. }

  
  dup CPa; rrdes CPa; clean; s in CA; apply HP in CA.
  s in CA; clear HP ST; des.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.
  -
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - unfold stack_top.
        rewrite LST -!pmplus_assoc pmplus_swap.
        prep_store; eauto.
    }
    { depmcplus0; s.
      rewrite LST -!pmplus_assoc pmplus_swap in SV0.
      eauto using pmvalid_plus_cell.
    }

    eapply dance_O; instantiate (1 := fun _ _ => depmach0); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      setcrf mrf.
      set (s7' := ({r1 r_sp - 1 |-> vax} ## fst s4, snd s7)).
      set (s'' := (mrf, P1, env, s6, s7')).
      assert (EXT': ext (wlift mylw) s'' (r1, P1, env, s6, s7)).
      { unfold s''; repeat (split; auto); unfold mrf; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }
      hexploit (KSIM g').
      * eauto.
      * eauto.
      * instantiate (1:=s'').
        eapply ext_trans; [eapply extpub_in_ext|]; eauto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * eapply vclos_mon; eauto using (gk_mon GK').
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            instantiate (1 := {| cpc := Nothing; crf := Just mrf
                               ; cst := freest; chp := hp |}).
            unfold mrf; rrdes CPa0; rr; esplits; eauto.
            s; tmext_simpl; eauto.
          - dup EXT'; rrdes EXT'.
            unfold s'' in *.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eauto using (gk_mon GK').
            + s; i; eapply (gk_mon GK'); eauto; split; eauto.
        }
        simpl cplus; noemp; rewrite -!pmplus_assoc (pmplus_swap freest).
        esplits; last first; eauto.
    + rr; esplits; s; eauto.
Qed.



Lemma tr_exp_correct_snd: forall F G ty ty' e
    (IH: ts_psim F G (Tprod ty ty') (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G ty' (fun n => map enc (tr_exp (Esnd e) n (map fst F) (map fst G))) (Esnd e).
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  eapply IH with (kb:= SKsnd::kb); eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  i; ts_prepare.

  invc VR REL; [|by inv VR]; rrdes CB; invc SVQ.
  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; s; noemp; rr; esplits; eauto. }

  
  dup CPa; rrdes CPa; clean; s in CA; apply HP in CA.
  s in CA; clear HP ST; des.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.
  -
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - unfold stack_top.
        rewrite LST -!pmplus_assoc pmplus_swap.
        prep_store; eauto.
    }
    { depmcplus0; s.
      rewrite LST -!pmplus_assoc pmplus_swap in SV0.
      eauto using pmvalid_plus_cell.
    }

    eapply dance_O; instantiate (1 := fun _ _ => depmach0); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      setcrf mrf.
      set (s7' := ({r1 r_sp - 1 |-> vay} ## fst s4, snd s7)).
      set (s'' := (mrf, P1, env, s6, s7')).
      assert (EXT': ext (wlift mylw) s'' (r1, P1, env, s6, s7)).
      { unfold s''; repeat (split; auto); unfold mrf; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }
      hexploit (KSIM g').
      * eauto.
      * eauto.
      * instantiate (1:=s'').
        eapply ext_trans; [eapply extpub_in_ext|]; eauto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * eapply vclos_mon; eauto using (gk_mon GK').
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            instantiate (1 := {| cpc := Nothing; crf := Just mrf
                               ; cst := freest; chp := hp |}).
            unfold mrf; rrdes CPa0; rr; esplits; eauto.
            s; tmext_simpl; eauto.
          - dup EXT'; rrdes EXT'.
            unfold s'' in *.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eauto using (gk_mon GK').
            + s; i; eapply (gk_mon GK'); eauto; split; eauto.
        }
        simpl cplus; noemp; rewrite -!pmplus_assoc (pmplus_swap freest).
        esplits; last first; eauto.
    + rr; esplits; s; eauto.
Qed.



Lemma tr_exp_correct_inl: forall F G ty ty' e
    (IH: ts_psim F G ty (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G (Tsum ty ty') (fun n => map enc (tr_exp (Einl e) n (map fst F) (map fst G))) (Einl e).
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ (wfo_nat 7)); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKinl::kb); eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  i; ts_prepare.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.
  -
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
      readpc EXT0 Eladdr laddr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - rewrite 2!pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
        prep_store; eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - unfold stack_top; rewrite LST -!pmplus_assoc pmplus_swap.
        prep_store; eauto.
    }
    { s; noemp; pmcells_fold; depmcplus0. 
      rewrite LST -!pmplus_assoc pmplus_swap in SV0.
      eauto using pmvalid_plus_cell.
    }

    eapply dance_O.
    instantiate (1 := fun _ cf _ _ _ _ _ => depmach1 cf); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.
      set (s7' := ({r1 r_sp - 1 |-> newloc} ## fst s4, snd s7)).
      set (s'' := (mrf, extend P1 (ts_inl va) newloc, env, s6, s7')).
      assert (EXT': ext (wlift mylw) s'' (r1, P1, env, s6, s7)).
      { unfold s'', extend; repeat (split; auto); unfold mrf; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }
      hexploit (KSIM g').
      * eauto.
      * eauto.
      * eapply ext_trans; eauto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * econs; eapply vclgen_inl; s; unfold extend; eauto; last first.
        { eapply vclos_mon; eauto using (gk_mon GK'). }
        { rr; splits; eauto.
          s; eapply ts_gw_vclos_closed; eauto.
        }
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            eapply tgt_confpred_extend; eauto.
            + insts $ {newloc |=> [0; va]}.
              split; eapply pmread_plus; eauto
              ; rewrite !pmcells_cons; s; pmplus_eq'.
            + unfold mrf; s; tmext_simpl; auto.
            + decomp_pmvalid HV; auto.
          - s; dup EXT'; rrdes EXT'.
            unfold s'' in *.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eauto using (gk_mon GK').
            + s; i; eapply (gk_mon GK'); eauto; split; eauto.
        }
        simpl cplus; noemp; rewrite -!pmplus_assoc (pmplus_swap freest).
        esplits; last first; eauto.
    + rr; esplits; s; eauto.
Qed.



Lemma tr_exp_correct_inr: forall F G ty ty' e
    (IH: ts_psim F G ty' (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G (Tsum ty ty') (fun n => map enc (tr_exp (Einr e) n (map fst F) (map fst G))) (Einr e).
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ (wfo_nat 7)); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKinr::kb); eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  i; ts_prepare.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.
  -
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
      readpc EXT0 Eladdr laddr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl; s.
        eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - rewrite 2!pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
        prep_store; eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - unfold stack_top; rewrite LST -!pmplus_assoc pmplus_swap.
        prep_store; eauto.
    }
    { s; noemp; pmcells_fold; depmcplus0. 
      rewrite LST -!pmplus_assoc pmplus_swap in SV0.
      eauto using pmvalid_plus_cell.
    }

    eapply dance_O.
    instantiate (1 := fun _ cf _ _ _ _ _ => depmach1 cf); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.
      set (s7' := ({r1 r_sp - 1 |-> newloc} ## fst s4, snd s7)).
      set (s'' := (mrf, extend P1 (ts_inr va) newloc, env, s6, s7')).
      assert (EXT': ext (wlift mylw) s'' (r1, P1, env, s6, s7)).
      { unfold s'', extend; repeat (split; auto); unfold mrf; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }
      hexploit (KSIM g').
      * eauto.
      * eauto.
      * eapply ext_trans; eauto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * econs; eapply vclgen_inr; s; unfold extend; eauto; last first.
        { eapply vclos_mon; eauto using (gk_mon GK'). }
        { rr; splits; eauto.
          s; eapply ts_gw_vclos_closed; eauto.
        }
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            eapply tgt_confpred_extend; eauto.
            + insts $ {newloc |=> [1; va]}.
              exists 1; splits; first (by ii; vauto); split; eapply pmread_plus; eauto
              ; rewrite !pmcells_cons; s; pmplus_eq'.
            + unfold mrf; s; tmext_simpl; auto.
            + decomp_pmvalid HV; auto.
          - s; dup EXT'; rrdes EXT'.
            unfold s'' in *.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eauto using (gk_mon GK').
            + s; i; eapply (gk_mon GK'); eauto; split; eauto.
        }
        simpl cplus; noemp; rewrite -!pmplus_assoc (pmplus_swap freest).
        esplits; last first; eauto.
    + rr; esplits; s; eauto.
Qed.



Lemma tr_exp_correct_pair: forall F G ty ty' e e'
    (IH: ts_psim F G ty (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (IH': ts_psim F G ty' (fun n => map enc (tr_exp e' n (map fst F) (map fst G))) e'),
  ts_psim F G (Tprod ty ty') (fun n => map enc (tr_exp (Epair e e') n (map fst F) (map fst G))) (Epair e e').
Proof.
  i; rdes IH; rdes IH'; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKpairL _ ::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.

  destruct (LEM (pmvalid (snd (snd s')))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ (wfo_nat 8)); rewrite <-wfo_add_assoc.
  eapply IH' with (kb:= SKpairR _ ::kb); clear IH'; eauto using ext_trans.
  { i; apply RSV in IN; des; esplits; eauto.
    - rrdes EXT0 EXT0 EXT0 EXT0; rewrite EXT3; eauto.
    - eapply vclos_mon; eauto using (gk_mon GK).
  }
  { i; apply RDV in IN; des; esplits; eauto.
    eapply vclos_mon; eauto using (gk_mon GK).
  }
  { rewrite ENV0.
    rrdes EXT0 EXT0 EXT0 EXT0.
    rrdes EXT1; eapply reprlist_mon; eauto.
  }
  { omega. }
  { rrdes EXT0 EXT0 EXT0 EXT0; rrdes EXT1.
    hexploit (pmplus_from_read EXT1); eauto.
    instantiate (1 := length (map enc (tr_exp e n (map fst F) (map fst G))) + n).
    i; des; rewrite H Eladdr !map_app !pmcells_app plus_comm.
    econs; symmetry.
    replace (_ - _) with (laddr - n); last first.
    { clear; set (x := length _); omega. }
    rewrite <-!pmplus_assoc.
    rhs rewrite pmplus_swap. rewrite !map_length.
    rewrite ->pmplus_swap; pmplus_eq'.
  }
  
  i; ts_prepare.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.
  - assert (EXT2: ext _ s'0 (r0, P0, s, s3, s4))
      by (eapply ext_trans; eauto).

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc EXT2 Eladdr laddr.
        unfold pop; auto. 
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT2 Eladdr laddr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
      readpc EXT2 Eladdr laddr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT2 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        symmetry; rewrite pmplus_swap LST0 LST SP0 SP.
        rewrite -!pmplus_assoc pmplus_swap.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT2 Eladdr laddr.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc EXT2 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        symmetry; rewrite pmplus_swap LST0 LST SP0 SP.
        rewrite -!pmplus_assoc.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT2 Eladdr laddr.
      - rewrite 2!pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
        prep_store; eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT2 Eladdr laddr.
      - unfold stack_top; rewrite pmplus_swap LST0 LST -!pmplus_assoc pmplus_swap.
        match goal with |- context[{?x - 1 |-> va}] =>
          
          replace x with (fst (fst (fst (fst s'0))) r_sp - 1) by (apply eq_minus; auto) 
        end.
        prep_store; eauto.
    }
    { s; noemp; pmcells_fold; depmcplus0.
      rewrite LST0 LST -!pmplus_assoc pmplus_swap in SV0.
      rewrite pmplus_swap.
      eapply pmvalid_plus_cell.
      rewrite pmplus_swap.
      rewrite SP0 in SV0 *.
      rewrite (pmplus_swap freest) pmplus_swap (pmplus_swap freest)in SV0.
      repeat match goal with
        | |- context[?x + 1 - 1] => replace (x + 1 - 1) with x by omega
        | H: context[?x + 1 - 1] |- _ => replace (x + 1 - 1) with x in H by omega
      end.
      eauto using pmvalid_plus_cell.
    }

    eapply dance_O.
    instantiate (1 := fun _ _ cf _ _ _ _ _ => depmach1 cf); s; eauto.

  - clear Eladdr RSV RDV.
    destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    destruct s'0 as [[[[]]]].
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_dep 2 RLa XTRa cfas; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      clear RLa0.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.
      set (s10' := ({r2 r_sp - 1 - 1|-> newloc} ## fst s4, snd s10)).
      set (s'' := (mrf, extend P2 (ts_pair va va0) newloc, s8, s9, s10')).
      assert (EXT': ext (wlift mylw) s'' (r2, P2, s8, s9, s10)).
      { unfold s'', extend; repeat (split; auto); unfold mrf; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }
      hexploit (KSIM g'0); clear KSIM.
      * eauto.
      * eauto.
      * do 2 eapply ext_trans; eauto.
      * s; unfold mrf; tmext_simpl.
        rewrite ENV1; auto.
      * s; unfold mrf; tmext_simpl.
        rewrite SP0; omega.
      * s; unfold mrf; tmext_simpl; rewrite RET0 RET; auto.
      * s; unfold mrf; tmext_simpl; auto.
      * econs; eapply vclgen_pair; try (by s; unfold extend; eauto); last first.
        { eapply vclos_mon; eauto using (gk_mon GK'0). }
        { eapply vclos_mon; eauto using (gk_mon GK'0), ext_trans. }
        { rr; splits; eauto.
          s; i; des; eapply ts_gw_vclos_closed in FH; eauto.
        }
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            assert (XXX: tgt_confpred (mrf, P2)
                                      {| cpc := Nothing;
                                         crf := Just mrf;
                                         cst := {r2 r_sp - 1 |-> va0} ## freest;
                                         chp := hp |}).
            { rewrite pmcells_sng.
              replace (r2 r_sp - 1) with (mrf r_sp); last by (unfold mrf; tmext_simpl).
              eapply tgt_confpred_pop; try (unfold mrf; tmext_simpl); eauto.
              rewrite -pmcells_sng.
              decomp_pmvalid SV; auto.
            }
            eapply tgt_confpred_extend; eauto.
            + insts $ {newloc |=> [va; va0]}.
              split; eapply pmread_plus; eauto
              ; rewrite !pmcells_cons; s; pmplus_eq'.
            + decomp_pmvalid HV; auto.
          - s; dup EXT'; rrdes EXT'.
            unfold s'' in *.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eauto using (gk_mon GK'0).
            + s; i; eapply (gk_mon GK'0); eauto; split; eauto.
        }
        simpl cplus; intro ES; esplits; last first.
        { eapply pesim_rewrite; eauto; [inst; f_equal; noemp].
          - rewrite !map_length !app_length.
            repeat match goal with |- context[tr_exp ?a ?b ?c ?d] =>
              let x := fresh "tr" in set (x := tr_exp a b c d)
            end.
            rewrite (plus_comm (length tr)).
            s; f_equal; rewrite plus_assoc; auto.
          - rewrite -!pmplus_assoc (pmplus_swap freest) pmplus_swap.
            rewrite (pmplus_comm freest); auto.
          - rewrite -!pmplus_assoc; auto.
        }
        eauto.
    + noemp; rr; esplits; s; eauto.
Qed.



Lemma tr_exp_correct_bop: forall F G e e' op
    (IH: ts_psim F G Tnat (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (IH': ts_psim F G Tnat (fun n => map enc (tr_exp e' n (map fst F) (map fst G))) e'),
  ts_psim F G Tnat (fun n => map enc (tr_exp (Ebop op e e') n (map fst F) (map fst G))) (Ebop op e e').
Proof.
  i; rdes IH; rdes IH'; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKbopL _ _ ::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.

  destruct (LEM (pmvalid (snd (snd s')))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH' with (kb:= SKbopR _ _ ::kb); clear IH'; eauto using ext_trans.
  { i; apply RSV in IN; des; esplits; eauto.
    - rrdes EXT0 EXT0 EXT0 EXT0; rewrite EXT3; eauto.
    - eapply vclos_mon; eauto using (gk_mon GK).
  }
  { i; apply RDV in IN; des; esplits; eauto.
    eapply vclos_mon; eauto using (gk_mon GK).
  }
  { rewrite ENV0.
    rrdes EXT0 EXT0 EXT0 EXT0.
    rrdes EXT1; eapply reprlist_mon; eauto.
  }
  { omega. }
  { rrdes EXT0 EXT0 EXT0 EXT0; rrdes EXT1.
    hexploit (pmplus_from_read EXT1); eauto.
    instantiate (1 := length (map enc (tr_exp e n (map fst F) (map fst G))) + n).
    i; des; rewrite H Eladdr !map_app !pmcells_app plus_comm.
    econs; symmetry.
    replace (_ - _) with (laddr - n); last first.
    { clear; set (x := length _); omega. }
    rewrite <-!pmplus_assoc.
    rhs rewrite pmplus_swap. rewrite !map_length.
    rewrite ->pmplus_swap; pmplus_eq'.
  }
  
  i; ts_prepare.

  inv VR VR0.
  invc REL; [|by invc VR1].
  invc REL0; [|by invc VR1].
  rrdes CA CB CA0 CB0; subst n0 n1; invc SVQ SVQ0.

  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; s; noemp; rr; esplits; eauto. }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.
  - assert (EXT2: ext _ s'0 (r0, P0, s, s3, s4))
      by (eapply ext_trans; eauto).

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc EXT2 Eladdr laddr.
        unfold pop; auto. 
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc EXT2 Eladdr laddr.
      - econs; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        s; rewrite LST0 LST SP0 SP.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
        (rhs rewrite -!pmplus_assoc -!(pmplus_swap {r0 r_sp |-> _})); auto.
      - econs; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        s; rewrite LST0 LST SP0 SP.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
        (rhs rewrite -!pmplus_assoc pmplus_swap); auto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT2 Eladdr laddr.
      - unfold stack_top; rewrite LST0 LST SP0 -!pmplus_assoc.
        rewrite -!(pmplus_swap {_|->va}).
        replace (_ r_sp - 1) with (fst (fst (fst (fst s'0))) r_sp - 1 - 1)
        ; [|rewrite SP0; omega].
        prep_store; eauto.
    }
    { s; noemp; tmext_simpl.
      rewrite LST0 LST SP0 -!pmplus_assoc pmplus_swap in SV0.
      depmcplus0; eauto using pmvalid_plus_cell.
      rewrite SP0.
      repeat match goal with
        | |- context[?a + 1 - 1] => replace (a + 1 - 1) with a by omega
        | H: context[?a + 1 - 1] |- _ => replace (a + 1 - 1) with a in H by omega
      end; eauto.
      rewrite (pmplus_swap freest) in SV0.
      rewrite -!(pmplus_swap {_|->va0}).
      eapply pmvalid_plus_cell.
      rewrite pmplus_swap in SV0; rewrite pmplus_swap.
      eapply pmvalid_plus_cell; eauto.
    }

    eapply dance_O.
    instantiate (1 := fun _ _ => depmach0); s; eauto.

  - destruct s'0 as [[[[]]]]; clean.
    i; rrdes XTRb RLb; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.

      set (s'' := (mrf, P1, env, s6,
                   ({r1 r_sp - 1 - 1 |-> bosem op va va0} ## fst s4, snd s7))).

      assert (EXT'': ext (wlift mylw) s'' (r1, P1, env, s6, s7)).
      { clear -HV ; unfold s''; s; splits; tmext_simpl; auto. }
      exists s''; esplits; [exact EXT''|].

      hexploit (KSIM g'0).
      * eauto.
      * eauto.
      * repeat (eapply ext_trans; eauto).
      * s; unfold mrf; tmext_simpl; congruence.
      * s; unfold mrf; tmext_simpl.
        rewrite SP0 SP; omega.
      * s; unfold mrf; tmext_simpl; congruence.
      * s; unfold mrf; tmext_simpl; auto.
      * s; econs; econs; rr; eauto.
      * intro X; simpl (match _ with | _ => _ end) in X.
        noemp; rewrite map_app app_length in X; simpl length in X.
        rewrite (plus_comm (length _)) in X; s in X.
        unfold mrf in X; tmext_simplH X; exploit X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto].
            eapply tgt_confpred_pop with (popped := [va0]); eauto.
            + unfold mrf; tmext_simpl; auto.
            + rewrite -pmcells_sng; auto.
              unfold mrf; tmext_simpl.
              rewrite pmplus_swap SP !(pmplus_swap freest) in SV.
              rewrite SP0 SP.
              decomp_pmvalid SV; eauto using pmvalid_plus_cell.
          - dup EXT''; rrdes EXT''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp.
        i; eapply pesim_rewrite; eauto.
        f_equal.
        { f_equal; rewrite !map_length !app_length; s; omega. }
        { unfold mrf; tmext_simpl.
          rewrite SP0 SP -!pmplus_assoc; noemp.
          repeat match goal with |- context[?a + 1 - 1] =>
            replace (a + 1 - 1) with a by omega
          end.
          rewrite -!(pmplus_swap {_|->va0}) (pmplus_swap freest); auto.
        }
        { rewrite -!pmplus_assoc; auto. }
    + rr; esplits; s; eauto.
      rewrite -plus_n_O; auto.
Grab Existential Variables.
  exact (wfo_nat 42).
Qed.
